#! -*- coding: utf-8
import os.path as path
import torch
import pandas as pd
import numpy as np
from torch.utils.data import Dataset

__all__ = ["LstmDataset",
           "MamlDataset",
           "LinearDataset"]


class LstmDataset(Dataset):
    def __init__(self,
                 queryfile,
                 supportfile,
                 tag: str = "",
                 predict_freq: int = 1,
                 num_query: int = None,
                 num_support: int = 3,
                 support_seq_len: int = None,
                 query_seq_len: int = None,
                 num_ar_feature: int = None,
                 train: bool = False):
        super(LstmDataset, self).__init__()
        assert predict_freq > 0
        assert num_ar_feature is None or num_ar_feature > 0
        assert num_query is None or num_query > 0
        assert num_support is None or num_support >= 0
        assert support_seq_len is None or support_seq_len > 0
        assert query_seq_len is None or query_seq_len > 0

        self.__num_query__ = num_query
        self.__num_support__ = num_support
        self.__num_ar_feature__ = num_ar_feature

        self.__train__ = train  # 学習モード

        # queryファイル名からドメイン、カテゴリを抽出
        basename, _ = path.splitext(path.basename(queryfile))
        basename = basename.replace(".query", "").replace(".support", "")
        items = basename.split("_")
        self.__domain__ = items[0]
        if len(items) == 1:
            self.__category__ = ""
        else:
            self.__category__ = items[1]
        self.__tag__ = tag

        self.__predict_freq__ = predict_freq
        self.__datas__, self.__indexs__ = [], []
        self.__datas__.append(np.load(queryfile))
        self.__indexs__.append(np.arange(len(self.__datas__[0])))
        if self.__train__:
            self.__datas__.append(self.__datas__[0])
            self.__indexs__ = [
                self.__indexs__[0][self.num_support:],
                self.__indexs__[0][:self.num_support]]
        else:
            self.__datas__.append(np.load(supportfile)[:self.num_support])
            self.__indexs__.append(np.arange(len(self.__datas__[1])))

        if query_seq_len is not None and self.__datas__[0].shape[1] <= query_seq_len:
            self.__query_seq_len__ = None  # 時系列長より大きい値が設定されている場合、全時系列を使用する
        else:
            self.__query_seq_len__ = query_seq_len  # query時系列長
        if support_seq_len is not None and self.__datas__[1].shape[1] <= support_seq_len:
            self.__support_seq_len__ = None  # 時系列長より大きい値が設定されている場合、全時系列を使用する
        else:
            self.__support_seq_len__ = support_seq_len  # support時系列長

        # query, supportの時系列長を記録
        # self.__seq_lengths__ = [d.shape[1] for d in self.__datas__]
        self.__seq_limit_indexs__ = [0 if l is None else (d.shape[1] - l)
                                     for d, l in zip(self.__datas__,
                                                     [self.query_seq_length, self.support_seq_length])]
        if self.num_query is not None:
            self.__length__ = np.min([self.num_query, len(self.__indexs__[0])])
        else:
            self.__length__ = len(self.__indexs__[0])
        self.__start_seq_indexs__ = [np.tile(0, len(i))
                                     for i in self.__indexs__]
        self.__shuffle_start_seq_indexs__()

    @property
    def domain(self): return self.__domain__
    @property
    def category(self): return self.__category__
    @property
    def tag(self): return self.__tag__

    @property
    def predict_freq(self): return self.__predict_freq__
    @property
    def num_query(self): return self.__num_query__
    @property
    def num_support(self): return self.__num_support__
    @property
    def num_ar_feature(self): return self.__num_ar_feature__
    @property
    def support_seq_length(self): return self.__support_seq_len__
    @property
    def query_seq_length(self): return self.__query_seq_len__
    @property
    def train(self): return self.__train__

    def __get_full_seq_query__(self, index):
        return self.__datas__[0][self.__indexs__[0][index]]

    def __get_full_seq_support__(self):
        return self.__datas__[1][self.__indexs__[1]]

    def support(self):
        s = self.__get_full_seq_support__()[:, :, np.newaxis]  # 3階のテンソルに変換
        if self.support_seq_length is None:
            # 全supportデータを返す
            return torch.Tensor(s)
        else:
            support = []
            for start_idx, vec in zip(self.__start_seq_indexs__[1], s):
                support.append(
                    vec[start_idx:start_idx+self.support_seq_length])
            return torch.Tensor(support)

    def shuffle(self):
        if self.train:
            indexs = np.concatenate(self.__indexs__)
            np.random.shuffle(indexs)
            self.__indexs__ = [indexs[self.num_support:],
                               indexs[:self.num_support]]
            self.__shuffle_start_seq_indexs__()

    def __shuffle_start_seq_indexs__(self):
        for idx, (seq_len, indexs, limit) in enumerate(
                zip([self.__query_seq_len__, self.__support_seq_len__],
                    self.__indexs__,
                    self.__seq_limit_indexs__)):
            if seq_len is None:
                continue
            else:
                self.__start_seq_indexs__[idx] = np.random.randint(
                    0, limit + 1, len(indexs))

    def __len__(self): return self.__length__

    def __getitem__(self, index):
        tensors = torch.Tensor(self.__get_full_seq_query__(index))
        if self.query_seq_length is not None:
            idx = self.__start_seq_indexs__[0][index]
            tensors = tensors[idx:idx+self.query_seq_length]
        inputs = tensors[:-self.predict_freq]
        targets = tensors[(self.predict_freq + self.num_ar_feature - 1):]
        return inputs.view(-1, 1), targets.view(-1, 1)


class MamlDataset(LstmDataset):
    def __init__(self,
                 queryfile,
                 supportfile,
                 tag: str = "",
                 predict_freq: int = 1,
                 num_query: int = None,
                 num_support: int = 3,
                 support_seq_len: int = None,
                 query_seq_len: int = None,
                 num_ar_feature: int = None,
                 train: bool = False):
        super(MamlDataset, self).__init__(
            queryfile, supportfile,
            tag=tag,
            predict_freq=predict_freq,
            num_query=num_query,
            num_support=num_support,
            support_seq_len=support_seq_len,
            query_seq_len=query_seq_len,
            num_ar_feature=num_ar_feature,
            train=train)

    def support(self):
        s = super().support()
        inputs = s[:, :-self.predict_freq]
        targets = s[:, (self.predict_freq + self.num_ar_feature - 1):]
        return inputs, targets
        # return inputs.unsqueeze(2), targets.unsqueeze(2)


class LinearDataset(LstmDataset):
    def __init__(self,
                 queryfile,
                 supportfile,
                 tag: str = "",
                 predict_freq: int = 1,
                 num_query: int = None,
                 num_support: int = 3,
                 support_seq_len: int = None,
                 query_seq_len: int = None,
                 num_ar_feature: int = None,
                 train: bool = False):
        super(LinearDataset, self).__init__(
            queryfile, supportfile,
            tag=tag,
            predict_freq=predict_freq,
            num_query=num_query,
            num_support=num_support,
            support_seq_len=support_seq_len,
            query_seq_len=query_seq_len,
            num_ar_feature=num_ar_feature,
            train=train)
        # self.__freqs__ = []
        # for d, seq_len in zip(self.__datas__, [self.query_seq_length, self.support_seq_length]):
        #     if seq_len is None:
        #         self.__freqs__.append(d.shape[1] - self.num_ar_feature)
        #     else:
        #         self.__freqs__.append(seq_len - self.num_ar_feature)

    def support(self):
        s = super().support()
        inputs = s[:, :-self.predict_freq]
        targets = s[:, (self.predict_freq + self.num_ar_feature - 1):]
        # freq = self.__freqs__[1]
        inputs = inputs.as_strided(
            (inputs.shape[0], targets.shape[1], self.num_ar_feature),
            (inputs.shape[1], 1, 1))
        return inputs, targets

    def __getitem__(self, index):
        inputs, targets = super().__getitem__(index)
        # freq = self.__freqs__[0]
        inputs = inputs.as_strided(
            (targets.shape[0], self.num_ar_feature), (1, 1))
        return inputs, targets
