#! -*- coding: utf-8
import torch
import torch.nn as nn

__all__ = ["LstmModel",
           "DomainModel",
           "BidrectionalModel",
           "AttentionModel",
           "AttentionOnlyModel",
           "LinearModel",
           "NonLinearModel"]


class LstmModel(nn.Module):
    def __init__(self,
                 input_dim: int = 1, hidden_dim: int = 64,
                 num_layers: int = 1, output_dim: int = 1):
        """
        Parameters
        ----------
        input_dim : int
            入力データの特徴量次元数
        hidden_dim : int
            LSTMの隠れ層次元数（出力次元数）
        num_layers : int
            反復回数
        output_dim: int
            出力データの次元数
        """
        super(LstmModel, self).__init__()
        input_dim, hidden_dim, num_layers, output_dim = int(
            input_dim), int(hidden_dim), int(num_layers), int(output_dim)
        assert input_dim > 0
        self.input_dim, self.output_dim = input_dim, output_dim
        self.lstm = nn.LSTM(input_size=input_dim, hidden_size=hidden_dim,
                            num_layers=num_layers, batch_first=True)
        self.output = nn.Sequential(
            nn.Linear(hidden_dim, hidden_dim),
            nn.ReLU(inplace=False),
            nn.Linear(hidden_dim, output_dim)
        )

    def forward(self, x, hidden0=None):
        x, (hidden, c) = self.lstm(x, hidden0)
        o = self.output(x)
        return o, (hidden, c)


class DomainModel(LstmModel):
    def __init__(self,
                 input_dim: int = 1,
                 hidden_dim: int = 64,
                 num_layers: int = 1,
                 output_dim: int = 1,
                 domain_hidden_dim: int = 64,
                 domain_num_layers: int = 1,
                 domain_forward_unit00: int = 64,
                 domain_forward_unit01: int = 64,
                 domain_forward_unit02: int = 32,
                 domain_output_size: int = 16):
        self.domain_input_dim = input_dim
        super(DomainModel, self).__init__(input_dim=input_dim + domain_output_size,
                                          hidden_dim=hidden_dim, num_layers=num_layers, output_dim=output_dim)

        self.domain_lstm = nn.LSTM(input_size=self.domain_input_dim, hidden_size=domain_hidden_dim,
                                   num_layers=domain_num_layers, batch_first=True)
        self.domain_forward00 = nn.Sequential(
            nn.Linear(domain_hidden_dim, domain_forward_unit00),
            nn.ReLU(inplace=False),
            nn.Linear(domain_forward_unit00, domain_forward_unit00),
            nn.ReLU(inplace=False),
            nn.Linear(domain_forward_unit00, domain_forward_unit00))
        self.domain_forward01 = nn.Sequential(
            nn.Linear(domain_forward_unit00, domain_forward_unit01),
            nn.ReLU(inplace=False),
            nn.Linear(domain_forward_unit01, domain_forward_unit01),
            nn.ReLU(inplace=False),
            nn.Linear(domain_forward_unit01, domain_forward_unit01))
        # feed_forward01と同様
        self.domain_forward02 = nn.Sequential(
            nn.Linear(domain_forward_unit01, domain_forward_unit02),
            nn.ReLU(inplace=False),
            nn.Linear(domain_forward_unit02, domain_forward_unit02),
            nn.ReLU(inplace=False),
            nn.Linear(domain_forward_unit02, domain_output_size))

    def domain_forward(self, x, hidden0=None):
        x, (hidden, c) = self.domain_lstm(x, hidden0)
        x = self.domain_forward00(x)
        x = x.mean(1)  # 時系列方向に平均
        x = self.domain_forward01(x)
        x = x.mean(0)  # データ方向に平均
        x = self.domain_forward02(x).view(-1)  # 単一ベクトルに変形
        return x, (hidden, c)


class BidrectionalModel(LstmModel):
    def __init__(self,
                 input_dim: int = 1,
                 hidden_dim: int = 64,
                 num_layers: int = 1,
                 output_dim: int = 1,
                 domain_hidden_dim: int = 64,
                 domain_num_layers: int = 1,
                 domain_forward_unit00: int = 64,
                 domain_forward_unit01: int = 64,
                 domain_forward_unit02: int = 32,
                 domain_output_size: int = 16):
        self.domain_input_dim = input_dim
        super(BidrectionalModel, self).__init__(input_dim=input_dim + domain_output_size,
                                                hidden_dim=hidden_dim, num_layers=num_layers, output_dim=output_dim)
        self.domain_lstm = nn.LSTM(input_size=self.domain_input_dim, hidden_size=domain_hidden_dim,
                                   num_layers=domain_num_layers, batch_first=True, bidirectional=True)
        # self.domain_lstm_r = nn.LSTM(input_size=self.domain_input_dim, hidden_size=domain_hidden_dim,
        #                              num_layers=domain_num_layers, batch_first=True)
        self.domain_forward00 = nn.Sequential(
            nn.Linear(domain_hidden_dim*2, domain_forward_unit00),
            nn.ReLU(inplace=False),
            nn.Linear(domain_forward_unit00, domain_forward_unit00),
            nn.ReLU(inplace=False),
            nn.Linear(domain_forward_unit00, domain_forward_unit00))
        self.domain_forward01 = nn.Sequential(
            nn.Linear(domain_forward_unit00, domain_forward_unit01),
            nn.ReLU(inplace=False),
            nn.Linear(domain_forward_unit01, domain_forward_unit01),
            nn.ReLU(inplace=False),
            nn.Linear(domain_forward_unit01, domain_forward_unit01))
        # feed_forward01と同様
        self.domain_forward02 = nn.Sequential(
            nn.Linear(domain_forward_unit01, domain_forward_unit02),
            nn.ReLU(inplace=False),
            nn.Linear(domain_forward_unit02, domain_forward_unit02),
            nn.ReLU(inplace=False),
            nn.Linear(domain_forward_unit02, domain_output_size))

    def domain_forward(self, x, hidden0=None):
        hs, hidden = self.domain_lstm(x, hidden0)
        x = self.domain_forward00(hs)
        x = x.mean(1)  # 時系列方向に平均
        x = self.domain_forward01(x)
        x = x.mean(0)  # データ方向に平均
        x = self.domain_forward02(x).view(-1)  # 単一ベクトルに変形
        return x, hidden


class AttentionModel(nn.Module):
    def __init__(self,
                 input_dim: int = 1,
                 hidden_dim: int = 64,
                 num_layers: int = 1,
                 output_dim: int = 1,
                 domain_hidden_dim: int = 64,
                 domain_num_layers: int = 1,
                 domain_output_size: int = 16):
        super(AttentionModel, self).__init__()
        self.hidden_dim = hidden_dim

        self.lstm = nn.LSTM(input_size=input_dim, hidden_size=hidden_dim,
                            num_layers=num_layers, batch_first=True)
        self.domain_lstm = nn.LSTM(input_size=input_dim, hidden_size=domain_hidden_dim,
                                   num_layers=domain_num_layers, batch_first=True, bidirectional=True)

        self.K = nn.Linear(domain_hidden_dim*2, hidden_dim, bias=False)
        self.Q = nn.Linear(hidden_dim, hidden_dim, bias=False)
        self.V = nn.Linear(domain_hidden_dim*2, hidden_dim, bias=False)

        self.output = nn.Sequential(
            nn.Linear(hidden_dim*2, hidden_dim),
            nn.ReLU(inplace=False),
            nn.Linear(hidden_dim, output_dim)
        )

    def forward(self, x, hs, hidden0=None):
        r, (hidden, c) = self.lstm(x, hidden0)
        khs = self.K(hs).view(-1, self.hidden_dim).unsqueeze(0).unsqueeze(0)
        qr = self.Q(r).unsqueeze(2)
        vhs = self.V(hs).view(-1, self.hidden_dim)
        x = (torch.softmax(khs*qr, dim=2)*vhs).sum(2)
        x = torch.cat([x, r], dim=-1)
        o = self.output(x)
        return o, (hidden, c)

    def domain_forward(self, x, hidden0=None):
        hs, hidden = self.domain_lstm(x, hidden0)
        return hs, hidden


class AttentionOnlyModel(nn.Module):
    def __init__(self,
                 input_dim: int = 1,
                 hidden_dim: int = 64,
                 num_layers: int = 1,
                 output_dim: int = 1,
                 domain_hidden_dim: int = 64,
                 domain_num_layers: int = 1,
                 domain_output_size: int = 16):
        super(AttentionOnlyModel, self).__init__()
        self.hidden_dim = hidden_dim

        self.lstm = nn.LSTM(input_size=input_dim, hidden_size=hidden_dim,
                            num_layers=num_layers, batch_first=True)
        self.domain_lstm = nn.LSTM(input_size=input_dim, hidden_size=domain_hidden_dim,
                                   num_layers=domain_num_layers, batch_first=True, bidirectional=True)

        self.K = nn.Linear(domain_hidden_dim*2, hidden_dim, bias=False)
        self.Q = nn.Linear(hidden_dim, hidden_dim, bias=False)
        self.V = nn.Linear(domain_hidden_dim*2, hidden_dim, bias=False)

        self.output = nn.Sequential(
            nn.Linear(hidden_dim, hidden_dim),
            nn.ReLU(inplace=False),
            nn.Linear(hidden_dim, output_dim)
        )

    def forward(self, x, hs, hidden0=None):
        r, (hidden, c) = self.lstm(x, hidden0)
        khs = self.K(hs).view(-1, self.hidden_dim).unsqueeze(0).unsqueeze(0)
        qr = self.Q(r).unsqueeze(2)
        vhs = self.V(hs).view(-1, self.hidden_dim)
        x = (torch.softmax(khs*qr, dim=2)*vhs).sum(2)
        # x = torch.cat([x, r], dim=-1)
        o = self.output(x)
        return o, (hidden, c)

    def domain_forward(self, x, hidden0=None):
        hs, hidden = self.domain_lstm(x, hidden0)
        return hs, hidden


class LinearModel(nn.Module):
    def __init__(self, input_dim: int = 1, output_dim: int = 1):
        super(LinearModel, self).__init__()
        self.linear = nn.Sequential(
            nn.Linear(input_dim, output_dim)
        )

    def forward(self, x, hidden0=None):
        return self.linear(x), hidden0


class NonLinearModel(nn.Module):
    def __init__(self, input_dim: int = 1, output_dim: int = 1, hidden_dim: int = 32):
        super(NonLinearModel, self).__init__()
        self.linear = nn.Sequential(
            nn.Linear(input_dim, hidden_dim),
            nn.ReLU(inplace=True),
            nn.Linear(hidden_dim, output_dim),
        )

    def forward(self, x, hidden0=None):
        return self.linear(x), hidden0


class MamlAttentionModel(nn.Module):
    def __init__(self, input_dim: int = 1,
                 hidden_dim: int = 64, num_layers: int = 1, output_dim: int = 1,
                 domain_hidden_dim: int = 64, domain_num_layers: int = 1, domain_output_size: int = 16):
        super(MamlAttentionModel, self).__init__()
        self.hidden_dim = hidden_dim

        self.lstm = nn.LSTM(input_size=input_dim, hidden_size=hidden_dim,
                            num_layers=num_layers, batch_first=True)
        self.domain_lstm = nn.LSTM(input_size=input_dim, hidden_size=domain_hidden_dim,
                                   num_layers=domain_num_layers, batch_first=True, bidirectional=True)

        self.K = nn.Linear(domain_hidden_dim*2, hidden_dim, bias=False)
        self.Q = nn.Linear(hidden_dim, hidden_dim, bias=False)
        self.V = nn.Linear(domain_hidden_dim*2, hidden_dim, bias=False)

        self.output = nn.Sequential(
            nn.Linear(hidden_dim*2, hidden_dim),
            nn.ReLU(inplace=False),
            nn.Linear(hidden_dim, output_dim)
        )

    def forward(self, x, support, hidden0=[None, None]):
        hs, hidden_support = self.domain_lstm(support, hidden0[0])
        r, hidden = self.lstm(x, hidden0[1])
        khs = self.K(hs).view(-1, self.hidden_dim).unsqueeze(0).unsqueeze(0)
        qr = self.Q(r).unsqueeze(2)
        vhs = self.V(hs).view(-1, self.hidden_dim)
        x = (torch.softmax(khs*qr, dim=2)*vhs).sum(2)
        x = torch.cat([x, r], dim=-1)
        o = self.output(x)
        return o, (hidden_support, hidden)
