#! -*- coding: utf-8
import os
import os.path as path
from argparse import ArgumentParser
from collections import OrderedDict
from concurrent.futures import ProcessPoolExecutor
from glob import glob

import numpy as np
import pandas as pd

if os.name == "posix":
    import matplotlib
    matplotlib.use("Agg")
    import matplotlib.pyplot as plt
else:
    import matplotlib.pyplot as plt


def __plot__(outdir, domain, dataset, queryfile, predictfiles, methods=["attention", "maml", "cross_domain"], format="png", num=30):
    queries = np.load(queryfile)
    queries = queries[:num]

    plotdatas = [queries]
    errors = OrderedDict()
    errors["domain"] = domain
    errors["dataset"] = dataset
    errors["no"] = 0
    for f, method in zip(predictfiles, methods):
        p = np.load(f)[:num]
        assert len(p) == len(queries), "invalid data size: %s, %s %s" % (
            f, p.shape, queries.shape)
        if p.ndim > 2:
            p = p.squeeze(-(p.ndim-2))
        # predict結果の先頭にnanをパディング
        pad = np.tile(np.nan, (len(p), queries.shape[1] - p.shape[1]))
        vals = np.concatenate([pad, p], axis=-1)
        plotdatas.append(vals)
        e = np.nanmean((queries - vals)**2, axis=-1)
        errors[method] = e
    # グラフの描画範囲（y軸）を設定
    rate = 0.1
    labels = ["query"] + methods

    errorfile = path.join(outdir, "summary_%s.csv" % (dataset))
    errors = pd.DataFrame(errors)
    errors["no"] = errors.index
    errors["argmin"] = np.argmin(errors.loc[:, methods].values, axis=-1)
    errors["var"] = np.var(errors.loc[:, methods].values, axis=-1)
    errors.to_csv(errorfile, index=False)

    errors = errors.loc[errors["argmin"] == 0]
    errors.sort_values(by=[methods[0], "var"],
                       ascending=[True, False],
                       inplace=True)
    errors.reset_index(inplace=True)
    print("%s %s target: %s" % (domain, dataset, len(errors)))
    for rowno, row in errors.loc[:5].iterrows():
        ylims = []
        for i, label in enumerate(labels):
            vals = plotdatas[i][row.no]
            plt.plot(vals, label=label)
            ylims.append(vals)
        imgfile = path.join(outdir,
                            "%s_%s_%03d.%s" % (domain, dataset, row.no, format))
        ylims = (np.nanmin(ylims), np.nanmax(ylims))
        ylims = (ylims[0] - np.abs(ylims[0]*rate),
                 ylims[1] + np.abs(ylims[1]*rate))
        plt.xlim(0, queries.shape[1])
        plt.ylim(ylims)
        plt.axis("off")
        plt.savefig(imgfile)
        if rowno == 0:
            plt.legend()
            plt.savefig(path.join(outdir, "sample.%s" % format))
        plt.close()

        for i, label in enumerate(labels):
            vals = plotdatas[i][row.no]
            plt.plot(vals, label=label)
            imgfile = path.join(outdir,
                                "%s_%s_%03d_%s.%s" % (domain, dataset, row.no, label, format))
            plt.xlim(0, queries.shape[1])
            plt.ylim(ylims)
            plt.axis("off")
            plt.savefig(imgfile)
            plt.close()

    print("%s %s target: %s done" % (domain, dataset, len(errors)))


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("rootdir", type=str)
    parser.add_argument("outdir", type=str)
    parser.add_argument("--targets",
                        default=["target"],
                        nargs="+", type=str,
                        choices=["train", "validation", "target"])
    parser.add_argument("--methods", type=str, nargs="+",
                        default=["attention", "maml", "cross_domain"],
                        choices=["attention", "maml", "cross_domain"])
    parser.add_argument("--format", default="png", type=str,
                        choices=["png", "jpg", "svg", "eps"])
    parser.add_argument("-num", type=int, default=30)
    args = parser.parse_args()
    print(args)
    if not path.isdir(args.outdir):
        os.makedirs(args.outdir, exist_ok=True)

    basemethod = args.methods[0]
    executor = ProcessPoolExecutor()
    tasks = []

    # 対象とする全datasetの結果を一度探索して、ドメインのリストを作成する
    domainset = []
    basetarget = args.targets[0]
    errors = {}
    counts = {}

    for datasetdir in glob(path.join(args.rootdir, "dataset*")):
        dataset = path.basename(datasetdir)
        for predictfile in glob(path.join(datasetdir, basemethod, basetarget, "*.predict.npy")):
            # print(predictfile)
            parentdir = path.dirname(predictfile)
            basename = path.basename(predictfile)
            domain, _ = path.splitext(basename)
            domain = domain.replace(".predict", "")
            queryfile = path.join(
                parentdir, "%s.query.npy" % (domain))
            if not path.isfile(queryfile):
                queryfile = path.join(
                    parentdir, "%s.npy" % (domain))
                assert path.isfile(
                    queryfile), "file not found: %s" % queryfile
            if not domain in errors:
                errors[domain] = [0] * len(args.methods)
                counts[domain] = 0

            queries = np.load(queryfile)[:args.num]
            for i, method in enumerate(args.methods):
                p = path.join(datasetdir, method, basetarget, basename)
                assert path.isfile(p), "file not found: " % p
                p = np.load(p)[:args.num]
                q = queries[:, -p.shape[1]:].reshape(p.shape)
                errors[domain][i] += np.mean(np.mean((p-q)**2, axis=-1))
            counts[domain] += 1
            # print(errors)

    domains = []
    for domain, vals in errors.items():
        # print(domain, vals)
        # e = np.ndarray(vals)/counts[domain]
        idx = np.argmin(vals)
        # print(domain, vals, idx)
        if idx == 0:
            domains.append(domain)
    domains = set(domains)
    print(domains)

    for datasetdir in glob(path.join(args.rootdir, "dataset*")):
        dataset = path.basename(datasetdir)
        for target in args.targets:
            for predictfile in glob(path.join(datasetdir, basemethod, target, "*.predict.npy")):
                parentdir = path.dirname(predictfile)
                basename = path.basename(predictfile)
                domain, _ = path.splitext(basename)
                if domain in domains:
                    continue
                domain = domain.replace(".predict", "")
                queryfile = path.join(
                    parentdir, "%s.query.npy" % (domain))
                if not path.isfile(queryfile):
                    queryfile = path.join(
                        parentdir, "%s.npy" % (domain))
                    assert path.isfile(
                        queryfile), "file not found: %s" % queryfile

                predictfiles = []
                for method in args.methods:
                    p = path.join(datasetdir, method, target, basename)
                    assert path.isfile(p), "file not found: " % p
                    predictfiles.append(p)

                outdir = path.join(args.outdir, domain)
                os.makedirs(outdir, exist_ok=True)

                # def __plot__(outdir, domain, dataset, queryfile, predictfiles, methods=["attention", "maml", "cross_domain"], format="png"):
                tasks.append(executor.submit(__plot__, outdir, domain, dataset,
                                             queryfile,
                                             predictfiles,
                                             methods=args.methods, format=args.format, num=args.num))

    for task in tasks:
        task.result()

    with open(path.join(args.outdir, "summary.csv"), "wt") as f:
        headers = ["domain", "dataset", "no"]
        headers.extend(args.methods)
        headers.append("diff")
        f.write(",".join(headers))
        f.write("\n")
        for domaindir in glob(path.join(args.outdir, "*")):
            if not path.isdir(domaindir):
                continue
            print(domaindir)
            df = pd.concat([pd.read_csv(f)
                            for f in glob(path.join(domaindir, "summary_*.csv"))])
            df = df[df["argmin"] == 0].copy()
            df.drop(columns=["argmin", "var"], inplace=True)
            df.reset_index(inplace=True)
            df["diff"] = df.loc[:, args.methods[1:]].min(
                axis=1) - df.loc[:, args.methods[0]]
            row = df.loc[df["diff"].idxmax()]
            line = [row.domain, row.dataset, row.no]
            line.extend([row[m] for m in args.methods])
            line.append(row["diff"])
            f.write(",".join(map(str, line)))
            f.write("\n")
            basename_prefix = "%s_%s_%03d" % (row.domain, row.dataset, row.no)
            for src in glob(path.join(args.outdir, row.domain, "%s*" % (basename_prefix))):
                dest = path.join(args.outdir, path.basename(src))
                if path.isfile(dest):
                    os.remove(dest)
                os.link(src, dest)
