# -*- coding: utf-8
# http://hilinker.hatenablog.com/entry/2018/06/23/204910
import torch
import torch.nn as nn
from torch.optim import SGD
import math
import numpy as np


class Predictor(nn.Module):
    def __init__(self, inputDim, hiddenDim, outputDim):
        super(Predictor, self).__init__()

        self.rnn = nn.LSTM(input_size=inputDim,
                           hidden_size=hiddenDim,
                           batch_first=True)
        self.output_layer = nn.Linear(hiddenDim, outputDim)

    def forward(self, inputs, hidden0=None):
        output, (hidden, cell) = self.rnn(inputs, hidden0)
        # 入力が50時系列あるので、出力も50時系列。最後の１時系列のみ出力。
        output = self.output_layer(output[:, -1, :])
        return output

def mkDataSet(data_size, data_length=50, freq=60., noise=0.00):
    """
    params\n
    data_size : データセットサイズ\n
    data_length : 各データの時系列長\n
    freq : 周波数\n
    noise : ノイズの振幅\n
    returns\n
    train_x : トレーニングデータ（t=1,2,...,size-1の値)\n
    train_t : トレーニングデータのラベル（t=sizeの値）\n
    """
    train_x = []
    train_t = []

    for offset in range(data_size):
        train_x.append([[math.sin(2 * math.pi * (offset + i) / freq) +
                         np.random.normal(loc=0.0, scale=noise)] for i in range(data_length)])
        train_t.append([math.sin(2 * math.pi * (offset + data_length) / freq)])

    return np.array(train_x), np.array(train_t)


def mkRandomBatch(train_x, train_t, device, batch_size=10):
    """
    train_x, train_tを受け取ってbatch_x, batch_tを返す。
    """
    indexs = np.random.randint(0, len(train_x) - 1, batch_size)
    batch_x = train_x[indexs]
    batch_t = train_t[indexs]

    batch_x, batch_t = torch.tensor(batch_x, dtype=torch.float32).to(
        device), torch.tensor(batch_t, dtype=torch.float32).to(device)
    return batch_x, batch_t


def main():
    # torch.backends.cudnn.enabled = False
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    training_size = 10000
    test_size = 1000
    epochs_num = 10
    hidden_size = 5
    batch_size = 100

    train_x, train_t = mkDataSet(training_size, data_length=50)
    test_x, test_t = mkDataSet(test_size, data_length=50)
    print("train x=%s, y=%s" % (train_x.shape, train_t.shape))
    print("test x=%s, y=%s" % (test_x.shape, test_t.shape))

    model = Predictor(1, hidden_size, 1)
    model.to(device)
    criterion = nn.MSELoss()
    optimizer = SGD(model.parameters(), lr=0.01)

    for epoch in range(epochs_num):
        # training
        running_loss = 0.0
        training_accuracy = 0.0
        for i in range(int(training_size / batch_size)):
            optimizer.zero_grad()

            data, label = mkRandomBatch(train_x, train_t, device, batch_size)

            output = model(data)
            # print("batch data:", data.shape, label.shape, output.shape)

            loss = criterion(output, label)
            # print("loss: ", loss)
            loss.backward()
            optimizer.step()

            # running_loss += loss.data[0]
            running_loss += loss.cpu().item()
            if device.type == "cuda":
                training_accuracy += np.sum(
                    np.abs((output.cpu().data - label.cpu().data).numpy()) < 0.1)
            else:
                training_accuracy += np.sum(np.abs((output.data -
                                                    label.data).numpy()) < 0.1)

        # test
        test_accuracy = 0.0
        for i in range(int(test_size / batch_size)):
            offset = i * batch_size
            data = torch.tensor(
                test_x[offset:offset+batch_size], dtype=torch.float32, device=device)
            label = torch.tensor(
                test_t[offset:offset+batch_size], dtype=torch.float32, device=device)
            output = model(data, None)

            if device.type == "cuda":
                output = output.cpu()
                label = label.cpu()
            test_accuracy += np.sum(np.abs((output.data -
                                            label.data).numpy()) < 0.1)

        training_accuracy /= training_size
        test_accuracy /= test_size

        print('%d loss: %.3f, training_accuracy: %.5f, test_accuracy: %.5f' % (
            epoch + 1, running_loss, training_accuracy, test_accuracy))


if __name__ == '__main__':
    main()
