#! -*- coding: utf-8
import os
import os.path as path
from argparse import ArgumentParser
import numpy as np
from glob import iglob as glob
from tempfile import TemporaryFile
from concurrent.futures import ProcessPoolExecutor
import pandas as pd


def __save_data__(domain, datadir, outdir, delimiter="\t", sequence_length=None):
    datafile = path.join(datadir, "%s.tsv" % domain)
    npyfile = path.join(outdir, "%s" % domain)
    try:
        datas = np.loadtxt(datafile, delimiter="\t")
        datas = datas[:, 1:]
        mean = np.nanmean(datas)
        var = np.nanvar(datas)
        datas = (datas - mean) / var
        if type(sequence_length) == int and sequence_length > 1:
            datas = datas[:, :sequence_length]

        np.save(npyfile, datas)
        # print(npyfile, datas.shape)

        saved = np.load("%s.npy" % (npyfile))
        assert np.sum((datas - saved) ** 2) == 0.0
    except Exception as e:
        print(e)
        print("Failed create data %s" % datafile)


def __split_data__(domain, datadir, outdir, delimiter="\t", num_support=5, num_query=5, shuffle=False, sequence_length=None):
    datafile = path.join(datadir, "%s.tsv" % domain)
    supportfile = path.join(outdir, "%s.support" % domain)
    queryfile = path.join(outdir, "%s.query" % domain)
    try:
        datas = np.loadtxt(datafile, delimiter="\t")
        datas = datas[:, 1:]
        mean = np.nanmean(datas)
        var = np.nanvar(datas)
        datas = (datas - mean) / var
        if shuffle:
            np.random.shuffle(datas)
        supports, queries = datas[:num_support], datas[num_support:num_support+num_query]
        if type(sequence_length) == int and sequence_length > 1:
            supports = supports[:, :sequence_length]
            queries = queries[:, :sequence_length]

        assert len(supports) > 0 and len(queries) > 0
        np.save(supportfile, supports)
        np.save(queryfile, queries)

        # print(supportfile, supports.shape)
        # print(queryfile, queries.shape)
        tmp = np.load("%s.npy" % (supportfile))
        assert np.sum((supports - tmp) ** 2) == 0.0
        tmp = np.load("%s.npy" % (queryfile))
        assert np.sum((queries - tmp) ** 2) == 0.0
    except Exception as e:
        print(e)
        print("Failed create data %s, (%s, %s)" %
              (datafile, supportfile, queryfile))


if __name__ == "__main__":
    """ データをsupport/queryに分割する。
    supportは評価には使用しない。
    """
    parser = ArgumentParser()
    parser.add_argument("rootdir", type=str)
    parser.add_argument("--outdir", type=str, default=None)
    parser.add_argument("--train", type=int, default=80,
                        help="trainに使用するdomain数")
    parser.add_argument("--validation", type=int, default=10,
                        help="validationに使用するdomain数")
    parser.add_argument("--min_data_count", type=int,
                        default=100, help="最小データ数")
    parser.add_argument("--min_sequence", type=int, default=50, help="最小時系列長")
    parser.add_argument("--sequence_length", type=int,
                        default=None, help="学習・検証データとして使用する時系列長")
    parser.add_argument("--num_support", type=int,
                        default=5, help="supportに使用する時系列数")
    parser.add_argument("--num_query", type=int,
                        default=5, help="queryに使用する時系列数")
    parser.add_argument("--seed", type=int, default=None)
    args = parser.parse_args()
    if args.outdir is None:
        args.outdir = args.rootdir
    for dirname in ["train", "validation", "target"]:
        outdir = path.join(args.outdir, dirname)
        if not path.isdir(outdir):
            os.makedirs(outdir)

    datainfo = pd.read_csv(path.join(args.rootdir, "datainfo.csv"))
    # データ数 >= args.min_data_count, 時系列長 >= args.min_sequence, 全データにnanを含まないものを対象とする
    # args.min_data_count = 100, args.min_sequence = 50の場合、103ドメインが対象
    datainfo = datainfo.loc[(datainfo["count"] >= args.min_data_count) & (
        datainfo["sequence"] >= args.min_sequence) & (datainfo["nan_count"] == 0)]
    assert len(datainfo) > args.train + args.validation
    print("domein count %d. train %d domains, validation %d domains, target %d domains" % (
        len(datainfo), args.train, args.validation, len(datainfo) - args.train - args.validation))
    domains = datainfo["domain"].values
    # print(domains.shape)

    np.random.seed(args.seed)
    train_domains = np.random.choice(domains, args.train, replace=False)
    domains = np.setdiff1d(domains, train_domains)
    val_domains = np.random.choice(domains, args.validation, replace=False)
    target_domains = np.setdiff1d(domains, val_domains)
    del domains
    # print(train_domains)
    # print(val_domains)
    # print(target_domains)
    # print(train_domains.shape, np.asarray(set(train_domains)).shape,
    #       val_domains.shape, target_domains.shape)

    executor = ProcessPoolExecutor()
    tasks = []

    for domains, tag in zip([train_domains, val_domains], ["train", "validation"]):
        for domain in domains:
            outdir = path.join(args.outdir, tag)
            tasks.append(executor.submit(__save_data__, domain,
                                         args.rootdir, outdir, delimiter="\t", sequence_length=args.sequence_length))

    for domains, tag in zip([target_domains], ["target"]):
        for domain in domains:
            outdir = path.join(args.outdir, tag)
            tasks.append(executor.submit(__split_data__, domain,
                                         args.rootdir, outdir, delimiter="\t", num_support=args.num_support, num_query=args.num_query, sequence_length=args.sequence_length))

    for task in tasks:
        task.result()
