#! -*- coding: utf-8
import os
import os.path as path
from argparse import ArgumentParser
import numpy as np
from glob import iglob as glob
from tempfile import TemporaryFile
from concurrent.futures import ProcessPoolExecutor
import pandas as pd


def __save_data__(domain, datadir, outdir, delimiter="\t", sequence_length=None):
    datafile = path.join(datadir, "%s.tsv" % domain)
    npyfile = path.join(outdir, "%s" % (domain))
    try:
        datas = np.loadtxt(datafile, delimiter="\t")
        datas = datas[:, 1:]
        if type(sequence_length) == int and sequence_length > 1:
            datas = datas[:, :sequence_length]

        mean = np.nanmean(datas)
        var = np.nanstd(datas)
        datas = (datas - mean) / var
        print("domain: %s, mean: %f, var: %f" %
              (domain, np.mean(datas), np.var(datas)))

        np.save(npyfile, datas)

        saved = np.load("%s.npy" % (npyfile))
        assert np.sum((datas - saved) ** 2) == 0.0
    except Exception as e:
        print(e)
        print("Failed create data %s" % datafile)


def __split_data__(domain, datadir, outdir, delimiter="\t", num_support=5, num_query=5, shuffle=False, sequence_length=None):
    datafile = path.join(datadir, "%s.tsv" % domain)
    supportfile = path.join(outdir, "%s.support" % (domain))
    queryfile = path.join(outdir, "%s.query" % (domain))
    try:
        datas = np.loadtxt(datafile, delimiter="\t")
        datas = datas[:, 1:]

        if type(sequence_length) == int and sequence_length > 1:
            datas = datas[:, :sequence_length]

        if shuffle:
            np.random.shuffle(datas)
        datas = datas[:num_support+num_query]
        mean = np.nanmean(datas)
        # var = np.nanvar(datas)
        var = np.nanstd(datas)
        datas = (datas - mean) / var
        print("domain: %s , mean: %f, var: %f" %
              (domain, np.mean(datas), np.var(datas)))

        supports, queries = datas[:num_support], datas[num_support:num_support+num_query]

        assert len(supports) > 0 and len(queries) > 0
        np.save(supportfile, supports)
        np.save(queryfile, queries)

        tmp = np.load("%s.npy" % (supportfile))
        assert np.sum((supports - tmp) ** 2) == 0.0
        tmp = np.load("%s.npy" % (queryfile))
        assert np.sum((queries - tmp) ** 2) == 0.0
    except Exception as e:
        print(e)
        print("Failed create data %s, (%s, %s)" %
              (datafile, supportfile, queryfile))


if __name__ == "__main__":
    """ データをsupport/queryに分割する。
    supportは評価には使用しない。
    """
    parser = ArgumentParser()
    parser.add_argument("rootdir", type=str)
    parser.add_argument("--outdir", type=str, default=None)
    parser.add_argument("--validation", type=int, default=25,
                        help="validationに使用するdomain数")
    parser.add_argument("--target", type=int, default=10,
                        help="targetに使用するdomain数")
    parser.add_argument("--min_data_count", type=int,
                        default=100, help="最小データ数")
    parser.add_argument("--min_sequence", type=int, default=20, help="最小時系列長")
    parser.add_argument("--sequence_length", type=int,
                        default=None, help="学習・検証データとして使用する時系列長")
    parser.add_argument("--num_support", type=int,
                        default=5, help="supportに使用する時系列数")
    parser.add_argument("--num_query", type=int,
                        default=5, help="queryに使用する時系列数")
    parser.add_argument("--seed", type=int, default=None)
    args = parser.parse_args()
    if args.outdir is None:
        args.outdir = args.rootdir
    for dirname in ["train", "validation", "target"]:
        outdir = path.join(args.outdir, dirname)
        if not path.isdir(outdir):
            os.makedirs(outdir)

    datainfo = pd.read_csv(
        path.join(args.rootdir, "datainfo_categories.csv"))
    datainfo = datainfo.drop("category", axis=1).groupby("domain").agg(
        {"count": np.sum, "nan_count": np.sum, "sequence": np.mean, "max_seq": np.max, "min_seq": np.min})
    datainfo.reset_index(inplace=True)
    # データ数 >= args.min_data_count, 時系列長 >= args.min_sequence, 全データにnanを含まないものを対象とする
    # count >= 100, nan_count == 0, sequence >= 100 -> 86ドメイン
    datainfo = datainfo.loc[(datainfo["count"] >= args.min_data_count) & (
        datainfo["sequence"] >= args.min_sequence) & (datainfo["nan_count"] == 0)]
    assert len(datainfo) > args.validation + args.target
    print("domein count %d. train %d domains, validation %d domains, target %d domains" % (
        len(datainfo), len(datainfo) - args.target - args.validation, args.validation, args.target))
    datainfo.reset_index(drop=True, inplace=True)
    indexs = np.arange(len(datainfo))

    np.random.seed(args.seed)
    val_indexs = np.random.choice(indexs, args.validation, replace=False)
    indexs = np.setdiff1d(indexs, val_indexs)
    target_indexs = np.random.choice(indexs, args.target, replace=False)
    train_indexs = np.setdiff1d(indexs, target_indexs)
    print("domain: %d(train: %d,validation: %d, target: %d)"%(len(datainfo), len(train_indexs), len(val_indexs), len(target_indexs)))
    del indexs

    executor = ProcessPoolExecutor()
    tasks = []

    for indexs, tag in zip([train_indexs], ["train"]):
        df = datainfo.iloc[indexs]
        for index, row in df.iterrows():
            outdir = path.join(args.outdir, tag)
            tasks.append(executor.submit(__save_data__,
                                         row.domain,
                                         args.rootdir,
                                         outdir,
                                         delimiter="\t",
                                         sequence_length=args.sequence_length))

    for indexs, tag in zip([val_indexs, target_indexs], ["validation", "target"]):
        df = datainfo.iloc[indexs]
        for index, row in df.iterrows():
            outdir = path.join(args.outdir, tag)
            tasks.append(executor.submit(__split_data__, row.domain,
                                         args.rootdir,
                                         outdir,
                                         delimiter="\t",
                                         num_support=args.num_support,
                                         num_query=args.num_query,
                                         sequence_length=args.sequence_length))

    for task in tasks:
        task.result()
