#! -*- coding: utf-8
import abc
import os
import sys
from datetime import datetime
from glob import glob
from logging import getLogger
from os import path

import numpy as np
import torch
from torch.utils.data import DataLoader

from datasets import SimpleCustomBatch as Batcher
from datasets import LstmDataset as Dataset

__all__ = ["AbstractTrainer"]


class AbstractTrainer(metaclass=abc.ABCMeta):
    def __init__(self,
                 predict_freq: int = 1,
                 num_query: int = 50,
                 num_train_support: int = 0,
                 num_predict_support: int = 0,
                 support_seq_len: int = 5,
                 query_seq_len: int = 5,
                 num_ar_feature: int = 3,
                 use_cuda: bool = False,
                 batch_size: int = 50,
                 workers: int = 1,
                 delimiter: str = "\t",
                 **kwargs):
        super(AbstractTrainer, self).__init__()
        self.logger = getLogger(__name__)
        self.__trains__, self.__vals__, self.__tests__ = [], [], []
        self.__predict_files__ = [self.trains, self.validations, self.tests], [
            "train", "validation", "target"]
        self.predict_freq = predict_freq
        self.num_query = num_query
        self.num_train_support = num_train_support
        self.num_predict_support = num_predict_support
        self.support_seq_len = support_seq_len
        self.query_seq_len = query_seq_len
        self.num_ar_feature = num_ar_feature

        self.use_cuda = use_cuda
        self.batch_size = batch_size
        self.workers = workers

        self.delimiter = delimiter

    @property
    def trains(self): return self.__trains__
    @property
    def validations(self): return self.__vals__
    @property
    def tests(self): return self.__tests__
    @property
    def use_cuda(self): return self.__use_cuda__
    @property
    def device(self): return self.__device__
    @use_cuda.setter
    def use_cuda(self, value: bool):
        if value:
            if torch.cuda.is_available():
                self.__use_cuda__ = True
                self.__device__ = torch.device("cuda")
            else:
                raise ValueError(
                    "CUDA is not available. YOU set use_cuda = False.")
        else:
            self.__use_cuda__ = False
            self.__device__ = torch.device("cpu")

    def create_datas(self, root_dir, outdir, train_domains: int = None, seed=None):
        for data_type, filelist in zip(["train"],  [self.trains]):
            target_files = np.asarray(
                glob(path.join(root_dir, data_type, "*.npy")))
            if train_domains is not None:
                np.random.seed(seed)
                indexs = np.arange(len(target_files))
                indexs = np.random.choice(indexs, train_domains, replace=False)
                target_files = target_files[indexs]

            for target_org_file in target_files:
                target_dir = path.join(outdir, data_type)
                target_file = path.join(
                    target_dir, path.basename(target_org_file))
                if not path.isdir(target_dir):
                    os.makedirs(target_dir)
                if not path.isfile(target_file) and not path.islink(target_file):
                    os.symlink(path.abspath(target_org_file), target_file)
                # ファイルリストに入力ファイルとsupportファイルのタプルを登録する
                # 学習時は入力ファイルのみを使用する。
                filelist.append((target_file, target_file))
        for data_type, filelist in zip(["validation", "target"],  [self.validations, self.tests]):
            for target_org_file in glob(path.join(root_dir, data_type, "*.query.npy")):
                support_org_file = target_org_file.replace(
                    ".query.npy", ".support.npy")
                target_dir = path.join(outdir, data_type)
                target_file = path.join(
                    target_dir, path.basename(target_org_file))
                support_file = path.join(
                    target_dir, path.basename(support_org_file))
                if not path.isdir(target_dir):
                    os.makedirs(target_dir)
                if not path.isfile(target_file) and not path.islink(target_file):
                    os.symlink(path.abspath(target_org_file), target_file)
                if not path.isfile(support_file) and not path.islink(support_file):
                    os.symlink(path.abspath(
                        support_org_file), support_file)
                # ファイルリストに入力ファイルとsupportファイルのタプルを登録する
                # 学習時は入力ファイルのみを使用する。
                filelist.append((target_file, support_file))

    def create_datas_specific(self, root_dir: str, outdir: str, domain: str, category: int, seed=None, num_support: int = 3):
        np.random.seed(seed)

        datadir = path.join(root_dir, "target")
        if category is None or len(category) == 0:
            basename = domain
        else:
            basename = "%s_%02d" % (domain, category)
        src_query = path.join(datadir, "%s.query.npy" % basename)
        src_support = path.join(datadir, "%s.support.npy" % basename)
        assert path.isfile(src_query) and path.isfile(src_support)

        # train
        target_dir = path.join(outdir, "train")
        dest_support = path.join(target_dir, "%s.support.npy" % basename)
        if not path.isdir(target_dir):
            os.makedirs(target_dir)
        if not path.isfile(dest_support):
            d = np.load(src_support)[:num_support]
            np.save(dest_support, d)
        self.trains.append((dest_support, dest_support))

        # target
        target_dir = path.join(outdir, "target")
        dest_query = path.join(target_dir, "%s.query.npy" % basename)
        dest_support = path.join(target_dir, "%s.support.npy" % basename)
        if not path.isdir(target_dir):
            os.makedirs(target_dir)
        if not path.isfile(dest_query) and not path.islink(dest_query):
            os.symlink(path.abspath(src_query), dest_query)
        if not path.isfile(dest_support) and not path.islink(dest_support):
            os.symlink(path.abspath(src_support), dest_support)
        self.tests.append((dest_query, dest_support))  # targetファイル

        for data_type, filelist in zip(["validation"],  [self.validations]):
            for target_org_file in glob(path.join(root_dir, data_type, "*.query.npy")):
                support_org_file = target_org_file.replace(
                    ".query.npy", ".support.npy")
                target_dir = path.join(outdir, data_type)
                target_file = path.join(
                    target_dir, path.basename(target_org_file))
                support_file = path.join(
                    target_dir, path.basename(support_org_file))
                if not path.isdir(target_dir):
                    os.makedirs(target_dir)
                if not path.isfile(target_file) and not path.islink(target_file):
                    os.symlink(path.abspath(target_org_file), target_file)
                if not path.isfile(support_file) and not path.islink(support_file):
                    os.symlink(path.abspath(
                        support_org_file), support_file)
                # ファイルリストに入力ファイルとsupportファイルのタプルを登録する
                # 学習時は入力ファイルのみを使用する。
                filelist.append((target_file, support_file))

    def Dataset(self, queryfile, supportfile,
                tag: str = "",
                num_query: int = None,
                num_support: int = 3,
                query_seq_len: int = None,
                train: bool = False,
                **kwargs):
        return Dataset(queryfile, supportfile,
                       tag=tag,
                       predict_freq=self.predict_freq,
                       num_query=num_query,
                       num_support=num_support,
                       support_seq_len=self.support_seq_len,
                       query_seq_len=query_seq_len,
                       num_ar_feature=self.num_ar_feature,
                       train=train)

    def DataLoaders(self, train: bool = False,
                    num_query: int = None,
                    num_support: int = 3,
                    query_seq_len: int = None,
                    **kwargs):
        loaders = []
        for files, t, tag in zip([self.trains, self.validations, self.tests],
                                 [train, False, False],
                                 self.__predict_files__[1]):
            dataloaders = []
            for query, support in files:
                dataset = self.Dataset(query, support,
                                       tag=tag,
                                       num_query=num_query,
                                       num_support=num_support,
                                       query_seq_len=query_seq_len,
                                       train=t, **kwargs)
                if self.use_cuda:
                    loader = DataLoader(dataset,
                                        batch_size=self.batch_size,
                                        num_workers=self.workers,
                                        collate_fn=lambda x: Batcher(x),
                                        pin_memory=True)
                else:
                    loader = DataLoader(dataset,
                                        batch_size=self.batch_size,
                                        num_workers=self.workers)
                loader.epoch = 0  # カスタムプロパティ追加
                dataloaders.append(loader)
            loaders.append(dataloaders)
        return loaders

    @abc.abstractmethod
    def Model(self, **kwargs): pass

    def Loss(self, **kwargs):
        reduction = kwargs["reduction"] if "reduction" in kwargs else "sum"
        # reduction = kwargs["reduction"] if "reduction" in kwargs else "mean"
        return torch.nn.MSELoss(reduction=reduction)

    def Optimizer(self, parameters, **kwargs):
        lr = kwargs["learning_rate"] if "learning_rate" in kwargs else 0.0001
        return torch.optim.Adam(parameters, lr=lr)

    def predict_file(self, sourcefile, data_type, outdir, ext="tsv", **kwargs):
        if ext is None or len(ext) == 0:
            ext = ""
        elif ext[0] == ".":
            pass
        else:
            ext = ".%s" % ext
        basename, _ = path.splitext(path.basename(sourcefile))
        dir = path.join(outdir, data_type)
        if not path.isdir(dir):
            os.makedirs(dir)
        basename = basename.replace("_normalized", "")
        targetfile = path.join(dir, "%s_%s%s" % (basename, "target", ext))
        predictfile = path.join(dir, "%s_%s%s" % (basename, "predict", ext))
        return targetfile, predictfile

    @property
    def predict_targets(self):
        """ 入力データのリストとscoreファイルのタグのTuple

        predictの結果はpredictメソッドで指定されたディレクトリ直下にタグ名のディレクトリを作成し、その直下に出力する
        """
        return self.__predict_files__

    @predict_targets.setter
    def predict_targets(self, value): self.__predict_files__ = value

    def train_batch(self,
                    model,
                    criterion,
                    loader,
                    **kwargs):
        total_loss, total_count = 0.0, 0
        if loader.dataset.train:
            loader.dataset.shuffle()
        domain = loader.dataset.domain
        category = loader.dataset.category
        tag = loader.dataset.tag
        loader.epoch += 1
        for batch_no, (inputs, targets) in enumerate(loader):
            start = datetime.now()

            inputs = inputs.to(self.device)
            targets = targets.to(self.device)

            # tmp1 = inputs[:,-targets.shape[1]:,-targets.shape[2]:].detach().cpu().numpy()
            # tmp2 = targets.detach().cpu().numpy()
            # tmp3 = np.sum((tmp1.ravel() - tmp2.ravel()) ** 2)
            # self.logger.log(5,"check %s %s %f"%(tmp1.shape, tmp2.shape, tmp3))
            # self.logger.log(
            #     5, "%s data shape=(inputs: %s, targets: %s)" %
            #     (tag, inputs.shape, targets.shape))

            # outputs, _ = model(inputs)
            outputs = model(inputs)
            if type(outputs) == tuple:
                outputs = outputs[0]
            self.logger.log(
                5, "%s data shape=(inputs: %s, outputs: %s, targets: %s)" %
                (tag, inputs.shape, outputs.shape, targets.shape))
            outputs = outputs[:, -targets.shape[1]:]

            loss = criterion(targets, outputs)
            count = np.prod(targets.shape)
            if loader.dataset.train:
                loss.backward(retain_graph=True)
            total_loss += loss.detach().item()
            total_count += count

            proc_time = (datetime.now() - start).total_seconds()
            self.logger.debug(
                "%s %04d.%03d %s.%s loss: %f count: %d proc time: %f" %
                (tag, loader.epoch, batch_no, domain, category,
                 total_loss/total_count, total_count, proc_time))
        return total_loss, total_count

    def predict_batch(self,
                      outdir,
                      model,
                      criterion,
                      loader,
                      **kwargs):
        total_loss, total_count = 0.0, 0
        domain = loader.dataset.domain
        category = loader.dataset.category
        tag = loader.dataset.tag
        loader.epoch += 1
        if category is None or len(category) == 0:
            outfile = path.join(outdir, "%s.predict.npy" % (domain))
        else:
            outfile = path.join(outdir,
                                "%s.%s.predict.npy" % (domain, category))

        predicts = []
        for batch_no, (inputs, targets) in enumerate(loader):
            start = datetime.now()

            inputs = inputs.to(self.device)
            targets = targets.to(self.device)
            # self.logger.log(
            #     5, "%s data shape=(inputs: %s, targets: %s)" %
            #     (tag, inputs.shape, targets.shape))

            outputs = model(inputs)
            if type(outputs) == tuple:
                outputs = outputs[0]
            self.logger.log(
                5, "%s data shape=(inputs: %s, outputs: %s, targets: %s)" %
                (tag, inputs.shape, outputs.shape, targets.shape))
            # outputs = outputs[:, (self.num_ar_feature-1):]
            outputs = outputs[:, -targets.shape[1]:]
            loss = criterion(targets, outputs)
            count = np.prod(targets.shape)

            total_loss += loss.detach().item()
            total_count += count
            # predictデータ保存用にリストに追加
            predicts.append(outputs.cpu().detach().numpy())

            proc_time = (datetime.now() - start).total_seconds()
            self.logger.debug(
                "%s %04d.%03d %s.%s loss: %f count: %d proc time: %f" %
                (tag, loader.epoch, batch_no, domain, category, total_loss/total_count, total_count, proc_time))

        predicts = np.concatenate(predicts, axis=0)  # batch次元で連結
        np.save(outfile, predicts)  # ファイルに保存

        return total_loss, total_count

    def train(self, modelfile: str,
              epochs: int = 10000,
              min_delta: float = 0.0,
              patience: int = 500,
              restore_best_weight: bool = False,
              update_interval: int = 1,
              seed: int = 7,
              **kwargs):
        ##### check args #####
        outdir = path.dirname(modelfile)
        if outdir is None or len(outdir) == 0:
            outdir = "."
        if not path.isdir(outdir):
            os.makedirs(outdir)

        if seed is not None:
            np.random.seed(seed)
            torch.manual_seed(seed)

        trainlogfile = kwargs["trainlog"] if "trainlog" in kwargs else path.join(
            outdir, "train.log")
        train_proclogfile = kwargs["train_proclog"] if "train_proclog" in kwargs else path.join(
            outdir, "train_proc.log")
        min_loss, best_epoch = sys.float_info.max, 0
        checkpoint = kwargs["checkpoint"] if "checkpoint" in kwargs else path.join(
            outdir, "checkpoint.pt")
        header = ["epoch"]

        dataloaders = self.DataLoaders(
            train=True,
            num_query=self.num_query,
            num_support=self.num_train_support,
            query_seq_len=self.query_seq_len)

        # tags = ["train", "validation", "target"]
        tags = self.__predict_files__[1]
        header.extend(["%s_loss" % tag for tag in tags])

        model = self.Model(**kwargs)
        model.to(self.device)
        criterion = self.Loss(**kwargs)
        optimizer = self.Optimizer(model.parameters(), **kwargs)

        with open(trainlogfile, "wt") as log, open(train_proclogfile, "wt") as proc_log:
            log.write(self.delimiter.join(header))
            log.write("\n")
            proc_log.write(
                self.delimiter.join(["epoch", "batch", "tag", "proc_time", "loss"]))
            proc_log.write("\n")

            for epoch in range(1, epochs + 1):
                epoch_start = datetime.now()

                counts = {tag: 0 for tag in tags}
                losses = {tag: 0.0 for tag in tags}

                for loaders, is_train, tag in zip(dataloaders, [True, False, False], tags):
                    if is_train:
                        model.train()
                    else:
                        model.eval()

                    optimizer.zero_grad()
                    for batch_no, loader in enumerate(loaders):
                        batch_start = datetime.now()

                        loss, count = self.train_batch(
                            model, criterion, loader, **kwargs)

                        losses[tag] += loss
                        counts[tag] += count
                        if is_train and (batch_no + 1) % update_interval == 0:
                            optimizer.step()
                            optimizer.zero_grad()

                        proc_time = (datetime.now() -
                                     batch_start).total_seconds()
                        proc_log.write(
                            self.delimiter.join(
                                map(str,
                                    [epoch, batch_no+1, tag,
                                     proc_time, losses[tag]/counts[tag]])))
                        proc_log.write("\n")

                    if is_train and (batch_no + 1) % update_interval != 0:
                        optimizer.step()
                        optimizer.zero_grad()

                proc_time = (datetime.now() - epoch_start).total_seconds()
                losses = {tag: losses[tag] / (1 if counts[tag] == 0 else counts[tag])
                          for tag in tags}
                self.logger.info("-%5d/%5d- loss: (%s) proc: %f sec" %
                                 (epoch, epochs,
                                  ", ".join(["%s=%f" % (i[0], i[1])
                                             for i in losses.items()]),
                                  proc_time))
                log_recode = [epoch]
                log_recode.extend([losses[tag] for tag in tags])
                log.write(self.delimiter.join(map(str, log_recode)))
                log.write("\n")
                if (min_loss - losses["validation"]) > min_delta:
                    torch.save(model.state_dict(), checkpoint)
                    min_loss = losses["validation"]
                    self.logger.debug(
                        "-%5d/%5d- update best score. save model" % (epoch, epochs))
                    best_epoch = epoch
                if (epoch - best_epoch) > patience:
                    if path.isfile(checkpoint) and restore_best_weight:
                        model.load_state_dict(torch.load(checkpoint))
                    self.logger.info(
                        "-%5d/%5d- early stoped and restore best weight: %s." % (epoch, epochs, restore_best_weight))
                    break
            # save model
            torch.save(model.state_dict(), modelfile)

    def predict(self,
                modelfile: str,
                outdir: str = None,
                **kwargs):
        ##### check args #####
        assert path.isfile(modelfile) or path.islink(
            modelfile), "model not fount: %s" % modelfile
        outrootdir = outdir
        if outrootdir is None:
            outrootdir = path.dirname(modelfile)
        elif len(outrootdir) == 0:
            outrootdir = "."

        if outrootdir is None or len(outrootdir) == 0:
            outrootdir = "."
        if not path.isdir(outrootdir):
            os.makedirs(outrootdir)

        model = self.Model(**kwargs)
        model.load_state_dict(torch.load(modelfile))
        model.to(self.device)
        model.eval()
        criterion = self.Loss(**kwargs)

        loaders = self.DataLoaders(
            train=True,
            num_query=None,
            num_support=self.num_predict_support,
            query_seq_len=None)
        tags = self.__predict_files__[1]

        score_file = path.join(outrootdir, "score.tsv")
        header = ["domain", "category", "type", "count", "error"]
        predict_proclogfile = kwargs["predict_proclog"] if "predict_proclog" in kwargs else path.join(
            outrootdir, "predict_proc.log")

        with open(score_file, "wt") as f, open(predict_proclogfile, "wt") as proc_log:
            f.write(self.delimiter.join(header))
            f.write("\n")
            proc_log.write(
                self.delimiter.join(["epoch", "batch", "tag", "proc_time", "loss"]))
            proc_log.write("\n")

            counts = {tag: 0 for tag in tags}
            losses = {tag: 0.0 for tag in tags}
            procs = {tag: 0.0 for tag in tags}

            epoch_start = datetime.now()
            for loaderset, tag in zip(loaders, tags):

                outdir = path.join(outrootdir, tag)
                if not path.isdir(outdir):
                    os.makedirs(outdir)

                for batch_no, loader in enumerate(loaderset):
                    batch_start = datetime.now()

                    loss, count = self.predict_batch(
                        outdir, model, criterion, loader, **kwargs)

                    proc_time = (datetime.now() -
                                 batch_start).total_seconds()

                    f.write(
                        self.delimiter.join(
                            map(str, [loader.dataset.domain,
                                      loader.dataset.category,
                                      tag,
                                      count,
                                      loss/count])))
                    f.write("\n")

                    losses[tag] += loss
                    counts[tag] += count
                    procs[tag] += proc_time

                    proc_log.write(
                        self.delimiter.join(
                            map(str, [0,
                                      batch_no+1,
                                      tag,
                                      proc_time,
                                      losses[tag]/counts[tag]])))
                    proc_log.write("\n")

            proc_time = (datetime.now() - epoch_start).total_seconds()
            losses = {tag: losses[tag] / (1 if counts[tag] == 0 else counts[tag])
                      for tag in tags}
            self.logger.info(
                "predict info: %s" %
                (",".join(["%s: (loss: %f, count: %d, proc: %f sec)" %
                            (tag, losses[tag], counts[tag], procs[tag])
                           for tag in tags])))
