#! -*- coding: utf-8
import os
import sys
from datetime import datetime
from os import path

import numpy as np
import torch

from .AbstractTrainer import AbstractTrainer

__all__ = ["AttentionTrainer", "AttentionOnlyTrainer"]


class AttentionTrainer(AbstractTrainer):
    def __init__(self,
                 predict_freq: int = 1,
                 num_query: int = 50,
                 num_train_support: int = 0,
                 num_predict_support: int = 0,
                 support_seq_len: int = 5,
                 query_seq_len: int = 5,
                 num_ar_feature: int = 3,
                 use_cuda: bool = False,
                 batch_size: int = 50,
                 workers: int = 1,
                 delimiter: str = "\t",
                 **kwargs):
        super(AttentionTrainer, self).__init__(
            predict_freq=predict_freq,
            num_query=num_query,
            num_train_support=num_train_support,
            num_predict_support=num_predict_support,
            support_seq_len=support_seq_len,
            query_seq_len=query_seq_len,
            num_ar_feature=num_ar_feature,
            use_cuda=use_cuda,
            batch_size=batch_size,
            workers=workers,
            delimiter=delimiter,
            **kwargs)

    def Model(self, **kwargs):
        """
        input_size: int
        hidden_size: int
        num_layers: int
        domain_input_size: int
        domain_hidden_size: int
        domain_num_layers: int
        """
        # assert "input_size" in kwargs and "hidden_size" in kwargs and "num_layers" in kwargs and "domain_output_size" in kwargs
        # # 元の時系列データの特徴量数に、ドメインの特徴量抽出用モデルの出力次元数を加えたものが入力データ
        assert "input_size" in kwargs
        assert "hidden_size" in kwargs
        assert "num_layers" in kwargs
        assert "domain_output_size" in kwargs
        assert "domain_hidden_size" in kwargs
        assert "domain_num_layers" in kwargs
        assert "domain_output_size" in kwargs
        from models import AttentionModel as Model
        # 元の時系列データの特徴量数に、ドメインの特徴量抽出用モデルの出力次元数を加えたものが入力データ
        input_size = int(kwargs["input_size"])
        hidden_size = int(kwargs["hidden_size"])
        num_layers = int(kwargs["num_layers"])
        domain_hidden_size = int(kwargs["domain_hidden_size"])
        domain_num_layers = int(kwargs["domain_num_layers"])
        domain_output_size = int(kwargs["domain_output_size"])
        return Model(input_size,
                     hidden_size,
                     num_layers,
                     output_dim=1,
                     domain_hidden_dim=domain_hidden_size,
                     domain_num_layers=domain_num_layers,
                     domain_output_size=domain_output_size)

    def train_batch(self,
                    model,
                    criterion,
                    loader,
                    **kwargs):
        total_loss, total_count = 0.0, 0
        if loader.dataset.train:
            loader.dataset.shuffle()
        domain = loader.dataset.domain
        category = loader.dataset.category
        tag = loader.dataset.tag
        loader.epoch += 1

        supports = loader.dataset.support().to(self.device)
        hs = model.domain_forward(supports)
        if type(hs) == tuple:
            hs = hs[0]

        for batch_no, (inputs, targets) in enumerate(loader):
            start = datetime.now()

            inputs = inputs.to(self.device)
            targets = targets.to(self.device)

            outputs = model(inputs, hs)
            if type(outputs) == tuple:
                outputs = outputs[0]
            self.logger.log(
                5, "%s data shape=(inputs: %s, supports: %s, outputs: %s, targets: %s)" %
                (tag, inputs.shape, supports.shape, outputs.shape, targets.shape))
            outputs = outputs[:, -targets.shape[1]:]

            loss = criterion(targets, outputs)
            count = np.prod(targets.shape)
            if loader.dataset.train:
                loss.backward(retain_graph=True)
            total_loss += loss.detach().item()
            total_count += count

            proc_time = (datetime.now() - start).total_seconds()
            self.logger.debug(
                "%s %04d.%03d %s.%s loss: %f count: %d proc time: %f" %
                (tag, loader.epoch, batch_no, domain, category,
                 total_loss/total_count, total_count, proc_time))
        return total_loss, total_count

    def predict_batch(self,
                      outdir,
                      model,
                      criterion,
                      loader,
                      **kwargs):
        total_loss, total_count = 0.0, 0
        domain = loader.dataset.domain
        category = loader.dataset.category
        tag = loader.dataset.tag
        loader.epoch += 1
        if category is None or len(category) == 0:
            outfile = path.join(outdir, "%s.predict.npy" % (domain))
        else:
            outfile = path.join(outdir,
                                "%s.%s.predict.npy" % (domain, category))

        supports = loader.dataset.support().to(self.device)
        hs = model.domain_forward(supports)
        if type(hs) == tuple:
            hs = hs[0]

        predicts = []
        for batch_no, (inputs, targets) in enumerate(loader):
            start = datetime.now()

            inputs = inputs.to(self.device)
            targets = targets.to(self.device)
            # self.logger.log(
            #     5, "%s data shape=(inputs: %s, targets: %s)" %
            #     (tag, inputs.shape, targets.shape))

            outputs = model(inputs, hs)
            if type(outputs) == tuple:
                outputs = outputs[0]
            self.logger.log(
                5, "%s data shape=(inputs: %s, supports: %s, outputs: %s, targets: %s)" %
                (tag, inputs.shape, supports.shape, outputs.shape, targets.shape))
            # outputs = outputs[:, (self.num_ar_feature-1):]
            outputs = outputs[:, -targets.shape[1]:]
            loss = criterion(targets, outputs)
            count = np.prod(targets.shape)

            total_loss += loss.detach().item()
            total_count += count
            # predictデータ保存用にリストに追加
            predicts.append(outputs.cpu().detach().numpy())

            proc_time = (datetime.now() - start).total_seconds()
            self.logger.debug(
                "%s %04d.%03d %s.%s loss: %f count: %d proc time: %f" %
                (tag, loader.epoch, batch_no, domain, category, total_loss/total_count, total_count, proc_time))

        predicts = np.concatenate(predicts, axis=0)  # batch次元で連結
        np.save(outfile, predicts)  # ファイルに保存

        return total_loss, total_count


class AttentionOnlyTrainer(AttentionTrainer):
    def __init__(self,
                 predict_freq: int = 1,
                 num_query: int = 50,
                 num_train_support: int = 0,
                 num_predict_support: int = 0,
                 support_seq_len: int = 5,
                 query_seq_len: int = 5,
                 num_ar_feature: int = 3,
                 use_cuda: bool = False,
                 batch_size: int = 50,
                 workers: int = 1,
                 delimiter: str = "\t",
                 **kwargs):
        super(AttentionOnlyTrainer, self).__init__(
            predict_freq=predict_freq,
            num_query=num_query,
            num_train_support=num_train_support,
            num_predict_support=num_predict_support,
            support_seq_len=support_seq_len,
            query_seq_len=query_seq_len,
            num_ar_feature=num_ar_feature,
            use_cuda=use_cuda,
            batch_size=batch_size,
            workers=workers,
            delimiter=delimiter,
            **kwargs)

    def Model(self, **kwargs):
        """
        input_size: int
        hidden_size: int
        num_layers: int
        domain_input_size: int
        domain_hidden_size: int
        domain_num_layers: int
        """
        # assert "input_size" in kwargs and "hidden_size" in kwargs and "num_layers" in kwargs and "domain_output_size" in kwargs
        # # 元の時系列データの特徴量数に、ドメインの特徴量抽出用モデルの出力次元数を加えたものが入力データ
        assert "input_size" in kwargs
        assert "hidden_size" in kwargs
        assert "num_layers" in kwargs
        assert "domain_output_size" in kwargs
        assert "domain_hidden_size" in kwargs
        assert "domain_num_layers" in kwargs
        assert "domain_output_size" in kwargs
        from models import AttentionOnlyModel as Model
        # 元の時系列データの特徴量数に、ドメインの特徴量抽出用モデルの出力次元数を加えたものが入力データ
        input_size = int(kwargs["input_size"])
        hidden_size = int(kwargs["hidden_size"])
        num_layers = int(kwargs["num_layers"])
        domain_hidden_size = int(kwargs["domain_hidden_size"])
        domain_num_layers = int(kwargs["domain_num_layers"])
        domain_output_size = int(kwargs["domain_output_size"])
        return Model(input_size,
                     hidden_size,
                     num_layers,
                     output_dim=1,
                     domain_hidden_dim=domain_hidden_size,
                     domain_num_layers=domain_num_layers,
                     domain_output_size=domain_output_size)
