#! -*- coding: utf-8
import os
import sys
from datetime import datetime
from os import path

import numpy as np
import torch

from ..AbstractTrainer import AbstractTrainer


class AbstractLongTermTrainer(AbstractTrainer):
    def __init__(self,
                 predict_freq: int = 1,
                 num_query: int = 50,
                 num_train_support: int = 0,
                 num_predict_support: int = 0,
                 support_seq_len: int = 5,
                 query_seq_len: int = 5,
                 num_ar_feature: int = 3,
                 use_cuda: bool = False,
                 batch_size: int = 50,
                 workers: int = 1,
                 delimiter: str = "\t",
                 input_seq: int = 10,
                 **kwargs):
        super(AbstractLongTermTrainer, self).__init__(
            predict_freq=predict_freq,
            num_query=num_query,
            num_train_support=num_train_support,
            num_predict_support=num_predict_support,
            support_seq_len=support_seq_len,
            query_seq_len=query_seq_len,
            num_ar_feature=num_ar_feature,
            use_cuda=use_cuda,
            batch_size=batch_size,
            workers=workers,
            delimiter=delimiter,
            **kwargs)
        self.input_sequence = input_seq

    @property
    def input_sequence(self): return self.__input_seq__
    @input_sequence.setter
    def input_sequence(self, value):
        assert value > 0
        self.__input_seq__ = value

    def __next_batch_iter_inputs__(self, inputs, outputs):
        return outputs

    def __inner_batch_iter__(self,
                             model, criterion, inputs, targets,
                             tag: str = "",
                             epoch: int = 0,
                             batch_no: int = 0,
                             domain: str = "",
                             category: str = "",
                             train: bool = False):
        hidden = None
        i = inputs[:, 0:self.input_sequence]
        outputs = torch.Tensor([]).to(self.device)

        iter_start = datetime.now()
        o = model(i, hidden0=hidden)
        if type(o) == tuple:
            hidden = o[1]
            o = o[0]
        outputs = torch.cat([outputs, o], dim=1)
        self.logger.log(
            5, "%s -%3d- data shape=(i: %s, o: %s), inputs: %s, outputs: %s, targets: %s" %
            (tag, self.input_sequence-1,
                list(i.shape), list(o.shape),
                list(inputs.shape), list(outputs.shape), list(targets.shape)))

        # 次の入力データ
        # start_idx = self.input_sequence + batch_iter
        i = self.__next_batch_iter_inputs__(i[:, -1:], o[:, -1:])

        for batch_iter in range(self.input_sequence, inputs.shape[1]):
            iter_start = datetime.now()

            # self.logger.log(
            #     5, "%s -%3d- data shape=(i: %s, o: %s), inputs: %s, outputs: %s, targets: %s" %
            #     (tag, batch_iter,
            #      list(i.shape), None,
            #      list(inputs.shape), list(outputs.shape), list(targets.shape)))

            o = model(i, hidden0=hidden)
            if type(o) == tuple:
                hidden = o[1]
                o = o[0]
            outputs = torch.cat([outputs, o], dim=1)
            self.logger.log(
                5, "%s -%3d- data shape=(i: %s, o: %s), inputs: %s, outputs: %s, targets: %s" %
                (tag, batch_iter,
                 list(i.shape), list(o.shape),
                 list(inputs.shape), list(outputs.shape), list(targets.shape)))

            # 次の入力データ
            # start_idx = self.input_sequence + batch_iter
            i = self.__next_batch_iter_inputs__(i[:, -1:], o[:, -1:])

        outputs = outputs[:, -targets.shape[1]:]
        loss = criterion(targets, outputs)
        count = np.prod(targets.shape)
        if train:
            loss.backward(retain_graph=True)

        return (loss.cpu().detach().numpy(),
                count,
                outputs.cpu().detach().numpy())

    def train_batch(self,
                    model,
                    criterion,
                    loader,
                    **kwargs):
        total_loss, total_count = 0.0, 0
        if loader.dataset.train:
            loader.dataset.shuffle()
        domain = loader.dataset.domain
        category = loader.dataset.category
        tag = loader.dataset.tag
        loader.epoch += 1

        for batch_no, (inputs, targets) in enumerate(loader):
            start = datetime.now()

            inputs = inputs.to(self.device)
            targets = targets.to(self.device)
            self.logger.log(
                5, "%s data shape=(inputs: %s, targets: %s)" %
                (tag, inputs.shape, targets.shape))

            loss, count, _ = self.__inner_batch_iter__(
                model, criterion, inputs, targets,
                tag=tag,
                epoch=loader.epoch,
                batch_no=batch_no,
                domain=domain,
                category=category,
                train=loader.dataset.train)
            total_loss += loss
            total_count += count

            # start_idx, hidden = 0, None
            # offset = inputs.shape[1] - targets.shape[1] + 1
            # i = inputs[:, 0:offset]  # 初回のみ数時系列長のデータを入力する
            # for batch_iter in range(targets.shape[1]):
            #     iter_start = datetime.now()

            #     t = targets[:, batch_iter:batch_iter+1]

            #     o = model(i, hidden0=hidden)
            #     if type(o) == tuple:
            #         hidden = o[1]
            #         o = o[0]
            #     self.logger.log(
            #         5, "%s data shape=(inputs: %s, outputs: %s, targets: %s)" %
            #         (tag, i.shape, o.shape, t.shape))
            #     o = o[:, -1:]

            #     loss = criterion(t, o)
            #     count = np.prod(t.shape)
            #     if loader.dataset.train:
            #         loss.backward(retain_graph=True)
            #     total_loss += loss.detach().item()
            #     total_count += count

            #     start_idx = offset+batch_iter
            #     i = self.__next_batch_iter_inputs__(i, o)

            #     proc_time = (datetime.now() - iter_start).total_seconds()
            #     self.logger.log(
            #         3,
            #         "%s %04d.%03d.%03d %s.%s loss: %f count: %d proc time: %f" %
            #         (tag, loader.epoch, batch_no, batch_iter,
            #          domain, category,
            #          loss/count, count, proc_time))

            proc_time = (datetime.now() - start).total_seconds()
            self.logger.debug(
                "%s %04d.%03d %s.%s loss: %f count: %d proc time: %f" %
                (tag, loader.epoch, batch_no, domain, category,
                 total_loss/total_count, total_count, proc_time))

        return total_loss, total_count

    def predict_batch(self,
                      outdir,
                      model,
                      criterion,
                      loader,
                      **kwargs):
        total_loss, total_count = 0.0, 0
        domain = loader.dataset.domain
        category = loader.dataset.category
        tag = loader.dataset.tag
        loader.epoch += 1
        if category is None or len(category) == 0:
            outfile = path.join(outdir, "%s.predict.npy" % (domain))
        else:
            outfile = path.join(outdir,
                                "%s.%s.predict.npy" % (domain, category))

        predicts = []
        for batch_no, (inputs, targets) in enumerate(loader):
            start = datetime.now()

            inputs = inputs.to(self.device)
            targets = targets.to(self.device)
            self.logger.log(
                5, "%s data shape=(inputs: %s, targets: %s)" %
                (tag, inputs.shape, targets.shape))

            # start_idx, hidden = 0, None
            # offset = inputs.shape[1] - targets.shape[1] + 1
            # i = inputs[:, 0:offset]  # 初回のみ数時系列長のデータを入力する

            # outputs = []
            # for batch_iter in range(targets.shape[1]):
            #     iter_start = datetime.now()

            #     t = targets[:, batch_iter:batch_iter+1]

            #     o = model(i, hidden0=hidden)
            #     if type(o) == tuple:
            #         hidden = o[1]
            #         o = o[0]
            #     self.logger.log(
            #         5, "%s data shape=(inputs: %s, outputs: %s, targets: %s)" %
            #         (tag, i.shape, o.shape, t.shape))
            #     o = o[:, -1:]

            #     loss = criterion(t, o)
            #     count = np.prod(t.shape)
            #     total_loss += loss.detach().item()
            #     total_count += count
            #     outputs.append(o)

            #     start_idx = offset+batch_iter
            #     # i = o
            #     i = self.__next_batch_iter_inputs__(i, o)

            #     proc_time = (datetime.now() - iter_start).total_seconds()
            #     self.logger.log(
            #         3,
            #         "%s %04d.%03d.%03d %s.%s loss: %f count: %d proc time: %f" %
            #         (tag, loader.epoch, batch_no, batch_iter,
            #          domain, category, loss/count, count, proc_time))

            # outputs = np.concatenate(outputs, axis=1)

            loss, count, outputs = self.__inner_batch_iter__(
                model, criterion, inputs, targets,
                tag=tag,
                epoch=loader.epoch,
                batch_no=batch_no,
                domain=domain,
                category=category,
                train=False)
            total_loss += loss
            total_count += count

            self.logger.log(
                5, "%s data shape=(inputs: %s, outputs: %s, targets: %s)" %
                (tag, inputs.shape, outputs.shape, targets.shape))
            # predictデータ保存用にリストに追加
            predicts.append(outputs)

            proc_time = (datetime.now() - start).total_seconds()
            self.logger.debug(
                "%s %04d.%03d %s.%s loss: %f count: %d proc time: %f" %
                (tag, loader.epoch, batch_no, domain, category,
                 total_loss/total_count, total_count, proc_time))

        predicts = np.concatenate(predicts, axis=0)  # batch次元で連結
        np.save(outfile, predicts)  # ファイルに保存

        return total_loss, total_count
