# * -*- coding: utf-8
import os
import sys
from datetime import datetime
from os import path

import higher
import numpy as np
import torch

from .AbstractLongTermTrainer import AbstractLongTermTrainer as AbstractTrainer

from datasets import MamlDataset as Dataset
from datasets import LinearDataset as LinearDataset

__all__ = ["LongTermMamlTrainer",
           "LongTermMamlLinearTrainer",
           "LongTermMamlNonLinearTrainer"]


class LongTermMamlTrainer(AbstractTrainer):
    def __init__(self,
                 predict_freq: int = 1,
                 num_query: int = 50,
                 num_train_support: int = 0,
                 num_predict_support: int = 0,
                 support_seq_len: int = 5,
                 query_seq_len: int = 5,
                 num_ar_feature: int = 3,
                 use_cuda: bool = False,
                 batch_size: int = 50,
                 workers: int = 1,
                 input_seq: int = 10,
                 delimiter: str = "\t",
                 **kwargs):
        super(LongTermMamlTrainer, self).__init__(
            predict_freq=predict_freq,
            num_query=num_query,
            num_train_support=num_train_support,
            num_predict_support=num_predict_support,
            support_seq_len=support_seq_len,
            query_seq_len=query_seq_len,
            num_ar_feature=num_ar_feature,
            use_cuda=use_cuda,
            batch_size=batch_size,
            workers=workers,
            input_seq=input_seq,
            delimiter=delimiter,
            **kwargs)

    def Model(self, **kwargs):
        """
        input_size: int
        hidden_size: int
        num_layers: int
        """
        assert "input_size" in kwargs
        assert "hidden_size" in kwargs
        assert "num_layers" in kwargs
        input_size = int(kwargs["input_size"])
        hidden_size = int(kwargs["hidden_size"])
        num_layers = int(kwargs["num_layers"])

        from models import LstmModel as Model
        return Model(input_size, hidden_size, num_layers)

    def InnerOptimizer(self, parameters, **kwargs):
        lr = kwargs["inner_learning_rate"] if "inner_learning_rate" in kwargs else 0.005
        return torch.optim.Adam(parameters, lr=lr)

    def InnerLoss(self, **kwargs):
        reduction = kwargs["reduction"] if "reduction" in kwargs else "sum"
        return torch.nn.MSELoss(reduction=reduction)

    def Dataset(self, queryfile, supportfile,
                tag: str = "",
                num_query: int = None,
                num_support: int = 3,
                query_seq_len: int = None,
                train: bool = False,
                **kwargs):
        return Dataset(queryfile, supportfile,
                       tag=tag,
                       predict_freq=self.predict_freq,
                       num_query=num_query,
                       num_support=num_support,
                       support_seq_len=self.support_seq_len,
                       query_seq_len=query_seq_len,
                       num_ar_feature=self.num_ar_feature,
                       train=train)

    def __inner_batch_iter__(self,
                             model, criterion, inputs, targets,
                             tag: str = "",
                             epoch: int = 0,
                             batch_no: int = 0,
                             domain: str = "",
                             category: str = "",
                             optimizer=None,
                             train: bool = False):
        hidden = None
        i = inputs[:, 0:self.input_sequence]

        iter_start = datetime.now()
        o = model(i, hidden0=hidden)
        if type(o) == tuple:
            hidden = o[1]
            o = o[0]
        outputs = o
        self.logger.log(
            5, "%s -%3d- data shape=(i: %s, o: %s), inputs: %s, outputs: %s, targets: %s" %
            (tag, self.input_sequence-1,
                list(i.shape), list(o.shape),
                list(inputs.shape), list(outputs.shape), list(targets.shape)))

        # 次の入力データ
        i = self.__next_batch_iter_inputs__(i[:, -1:], o[:, -1:])

        for batch_iter in range(self.input_sequence, inputs.shape[1]):
            iter_start = datetime.now()

            o = model(i, hidden0=hidden)
            if type(o) == tuple:
                hidden = o[1]
                o = o[0]
            outputs = torch.cat([outputs, o], dim=1)
            self.logger.log(
                5, "%s -%3d- data shape=(i: %s, o: %s), inputs: %s, outputs: %s, targets: %s" %
                (tag, batch_iter,
                 list(i.shape), list(o.shape),
                 list(inputs.shape), list(outputs.shape), list(targets.shape)))

            # 次の入力データ
            i = self.__next_batch_iter_inputs__(i[:, -1:], o[:, -1:])

        outputs = outputs[:, -targets.shape[1]:]
        loss = criterion(targets, outputs)
        count = np.prod(targets.shape)
        if train:
            if optimizer is None:
                loss.backward(retain_graph=True)
            else:
                optimizer.step(loss)

        return (loss.cpu().detach().numpy(),
                count,
                outputs.cpu().detach().numpy())

    def train_batch(self,
                    model,
                    criterion,
                    loader,
                    inner_iter: int = 5,
                    **kwargs):
        total_loss, total_count = 0.0, 0
        if loader.dataset.train:
            loader.dataset.shuffle()
        domain = loader.dataset.domain
        category = loader.dataset.category
        tag = loader.dataset.tag
        loader.epoch += 1

        inner_criterion = self.InnerLoss(**kwargs)
        inner_optimizer = self.InnerOptimizer(model.parameters(), **kwargs)

        supports = loader.dataset.support()

        with higher.innerloop_ctx(
            model, inner_optimizer, copy_initial_weights=False
        ) as (fnet, diffopt):
            inputs, targets = supports
            inputs = inputs.to(self.device)
            targets = targets.to(self.device)
            for i in range(1, inner_iter+1):
                start = datetime.now()

                loss, count, _ = self.__inner_batch_iter__(
                    fnet, inner_criterion, inputs, targets,
                    tag="%s.inner" % tag,
                    epoch=loader.epoch,
                    batch_no=0,
                    domain=domain,
                    category=category,
                    optimizer=diffopt,
                    train=True)

                proc_time = (datetime.now() - start).total_seconds()
                self.logger.log(
                    8,
                    "%s %04d.%03d.%02d %s.%s loss: %f count: %d proc time: %f" %
                    (tag, loader.epoch, 0, i,
                        domain, category,
                        loss/count, count,
                        proc_time))
            for batch_no, (inputs, targets) in enumerate(loader):

                start = datetime.now()

                inputs = inputs.to(self.device)
                targets = targets.to(self.device)

                loss, count, _ = self.__inner_batch_iter__(
                    fnet, criterion, inputs, targets,
                    tag=tag,
                    epoch=loader.epoch,
                    batch_no=batch_no,
                    domain=domain,
                    category=category,
                    optimizer=None,
                    train=loader.dataset.train)
                total_loss += loss
                total_count += count

                proc_time = (datetime.now() - start).total_seconds()
                self.logger.debug(
                    "%s %04d.%03d %s.%s loss: %f count: %d proc time: %f" %
                    (tag, loader.epoch, batch_no, domain, category,
                     total_loss/total_count, total_count, proc_time))
        return total_loss, total_count

    def predict_batch(self,
                      outdir,
                      model,
                      criterion,
                      loader,
                      inner_iter: int = 5,
                      **kwargs):
        total_loss, total_count = 0.0, 0
        domain = loader.dataset.domain
        category = loader.dataset.category
        tag = loader.dataset.tag
        loader.epoch += 1
        if category is None or len(category) == 0:
            outfile = path.join(outdir, "%s.predict.npy" % (domain))
        else:
            outfile = path.join(outdir,
                                "%s.%s.predict.npy" % (domain, category))

        inner_criterion = self.InnerLoss(**kwargs)
        inner_optimizer = self.InnerOptimizer(model.parameters(), **kwargs)

        supports = loader.dataset.support()

        predicts = []
        with higher.innerloop_ctx(
            model, inner_optimizer, copy_initial_weights=False
        ) as (fnet, diffopt):
            inputs, targets = supports
            inputs = inputs.to(self.device)
            targets = targets.to(self.device)
            for i in range(1, inner_iter+1):
                start = datetime.now()

                loss, count, _ = self.__inner_batch_iter__(
                    fnet, inner_criterion, inputs, targets,
                    tag="%s.inner" % tag,
                    epoch=loader.epoch,
                    batch_no=0,
                    domain=domain,
                    category=category,
                    optimizer=diffopt,
                    train=True)

                proc_time = (datetime.now() - start).total_seconds()
                self.logger.log(
                    8,
                    "%s %04d.%03d.%02d %s.%s loss: %f count: %d proc time: %f" %
                    (tag, loader.epoch, 0, i,
                        domain, category,
                        loss/count, count,
                        proc_time))

            for batch_no, (inputs, targets) in enumerate(loader):

                start = datetime.now()

                inputs = inputs.to(self.device)
                targets = targets.to(self.device)

                loss, count, outputs = self.__inner_batch_iter__(
                    fnet, criterion, inputs, targets,
                    tag=tag,
                    epoch=loader.epoch,
                    batch_no=batch_no,
                    domain=domain,
                    category=category,
                    optimizer=None,
                    train=False)
                total_loss += loss
                total_count += count

                # predictデータ保存用にリストに追加
                predicts.append(outputs)

                proc_time = (datetime.now() - start).total_seconds()
                self.logger.debug(
                    "%s %04d.%03d %s.%s loss: %f count: %d proc time: %f" %
                    (tag, loader.epoch, batch_no, domain, category, total_loss/total_count, total_count, proc_time))

        predicts = np.concatenate(predicts, axis=0)  # batch次元で連結
        np.save(outfile, predicts)  # ファイルに保存

        return total_loss, total_count


class LongTermMamlLinearTrainer(LongTermMamlTrainer):
    def __init__(self,
                 predict_freq: int = 1,
                 num_query: int = 50,
                 num_train_support: int = 0,
                 num_predict_support: int = 0,
                 support_seq_len: int = 5,
                 query_seq_len: int = 5,
                 num_ar_feature: int = 3,
                 use_cuda: bool = False,
                 batch_size: int = 50,
                 workers: int = 1,
                 input_seq: int = 10,
                 delimiter: str = "\t",
                 **kwargs):
        super(LongTermMamlLinearTrainer, self).__init__(
            predict_freq=predict_freq,
            num_query=num_query,
            num_train_support=num_train_support,
            num_predict_support=num_predict_support,
            support_seq_len=support_seq_len,
            query_seq_len=query_seq_len,
            num_ar_feature=num_ar_feature,
            use_cuda=use_cuda,
            batch_size=batch_size,
            workers=workers,
            input_seq=input_seq,
            delimiter=delimiter,
            **kwargs)

    def Model(self, **kwargs):
        """
        input_size: int
        """
        assert "input_size" in kwargs
        from models import LinearModel as Model
        input_size = self.num_ar_feature
        output_size = int(kwargs["input_size"])
        return Model(input_size, output_size)

    def Dataset(self, queryfile, supportfile,
                tag: str = "",
                num_query: int = None,
                num_support: int = 3,
                query_seq_len: int = None,
                train: bool = False,
                **kwargs):
        return LinearDataset(queryfile, supportfile,
                             tag=tag,
                             predict_freq=self.predict_freq,
                             num_query=num_query,
                             num_support=num_support,
                             support_seq_len=self.support_seq_len,
                             query_seq_len=query_seq_len,
                             num_ar_feature=self.num_ar_feature,
                             train=train)

    def __next_batch_iter_inputs__(self, inputs, outputs):
        """ 次回batch iterationで使用する入力データを作成する
        =====
        今回入力したデータの特徴量次元末尾に今回の出力データを追加。
        追加したデータ数分、特徴量次元の先頭データを除外する。

        =====
        inputs: tensor
            今回入力データ
        outputs: tensor
            今回出力データ
        =====
        next_inputs: tensor
            次回入力データ
        """
        i = torch.cat([inputs, outputs], dim=2)
        i = i[:, :, outputs.shape[-1]:]
        return i


class LongTermMamlNonLinearTrainer(LongTermMamlLinearTrainer):
    def __init__(self,
                 predict_freq: int = 1,
                 num_query: int = 50,
                 num_train_support: int = 0,
                 num_predict_support: int = 0,
                 support_seq_len: int = 5,
                 query_seq_len: int = 5,
                 num_ar_feature: int = 3,
                 use_cuda: bool = False,
                 batch_size: int = 50,
                 workers: int = 1,
                 input_seq: int = 10,
                 delimiter: str = "\t",
                 **kwargs):
        super(LongTermMamlNonLinearTrainer, self).__init__(
            predict_freq=predict_freq,
            num_query=num_query,
            num_train_support=num_train_support,
            num_predict_support=num_predict_support,
            support_seq_len=support_seq_len,
            query_seq_len=query_seq_len,
            num_ar_feature=num_ar_feature,
            use_cuda=use_cuda,
            batch_size=batch_size,
            workers=workers,
            input_seq=input_seq,
            delimiter=delimiter,
            **kwargs)

    def Model(self, **kwargs):
        """
        input_size: int
        hidden_size: int
        """
        assert "input_size" in kwargs and "hidden_size" in kwargs
        from models import NonLinearModel as Model
        input_size = self.num_ar_feature
        hidden_size = int(kwargs["hidden_size"])
        output_size = int(kwargs["input_size"])
        return Model(input_size, output_size, hidden_size)
