#! -*- coding: utf-8
import logging
import os
import sys
from datetime import datetime
from os import path

import numpy as np
import torch

from datasets import LinearDataset as LinearDataset
from datasets import LstmDataset as Dataset

from .AbstractLongTermTrainer import AbstractLongTermTrainer as AbstractTrainer

__all__ = ["LongTermMetaLearnTrainer",
           "LongTermBidirectionalTrainer",
           "LongTermAttentionTrainer"]


class LongTermMetaLearnTrainer(AbstractTrainer):
    def __init__(self,
                 predict_freq: int = 1,
                 num_query: int = 50,
                 num_train_support: int = 0,
                 num_predict_support: int = 0,
                 support_seq_len: int = 5,
                 query_seq_len: int = 5,
                 num_ar_feature: int = 3,
                 use_cuda: bool = False,
                 batch_size: int = 50,
                 workers: int = 1,
                 input_seq: int = 10,
                 delimiter: str = "\t",
                 **kwargs):
        super(LongTermMetaLearnTrainer, self).__init__(
            predict_freq=predict_freq,
            num_query=num_query,
            num_train_support=num_train_support,
            num_predict_support=num_predict_support,
            support_seq_len=support_seq_len,
            query_seq_len=query_seq_len,
            num_ar_feature=num_ar_feature,
            use_cuda=use_cuda,
            batch_size=batch_size,
            workers=workers,
            input_seq=input_seq,
            delimiter=delimiter,
            **kwargs)

    def Model(self, **kwargs):
        """
        input_size: int
        hidden_size: int
        num_layers: int
        domain_input_size: int
        domain_hidden_size: int
        domain_num_layers: int
        """
        # assert "input_size" in kwargs and "hidden_size" in kwargs and "num_layers" in kwargs and "domain_output_size" in kwargs
        # # 元の時系列データの特徴量数に、ドメインの特徴量抽出用モデルの出力次元数を加えたものが入力データ
        assert "input_size" in kwargs and "hidden_size" in kwargs and "num_layers" in kwargs and "domain_output_size" in kwargs
        assert "domain_hidden_size" in kwargs and "domain_num_layers" in kwargs and "domain_output_size" in kwargs
        assert "domain_forward_unit00" in kwargs and "domain_forward_unit01" in kwargs and "domain_forward_unit02" in kwargs
        # 元の時系列データの特徴量数に、ドメインの特徴量抽出用モデルの出力次元数を加えたものが入力データ
        from models import DomainModel as DomainModel
        input_size = int(kwargs["input_size"])
        hidden_size = int(kwargs["hidden_size"])
        num_layers = int(kwargs["num_layers"])
        domain_hidden_size = int(kwargs["domain_hidden_size"])
        domain_num_layers = int(kwargs["domain_num_layers"])
        domain_output_size = int(kwargs["domain_output_size"])
        domain_forward_unit00 = int(kwargs["domain_forward_unit00"])
        domain_forward_unit01 = int(kwargs["domain_forward_unit01"])
        domain_forward_unit02 = int(kwargs["domain_forward_unit02"])
        return DomainModel(input_size,
                           hidden_size,
                           num_layers,
                           output_dim=1,
                           domain_hidden_dim=domain_hidden_size,
                           domain_num_layers=domain_num_layers,
                           domain_output_size=domain_output_size,
                           domain_forward_unit00=domain_forward_unit00,
                           domain_forward_unit01=domain_forward_unit01,
                           domain_forward_unit02=domain_forward_unit02)

    def __inner_batch_iter__(self,
                             model, criterion, inputs, targets, supports,
                             tag: str = "",
                             epoch: int = 0,
                             batch_no: int = 0,
                             domain: str = "",
                             category: str = "",
                             train: bool = False):
        hidden = None
        i = inputs[:, 0:self.input_sequence]
        outputs = torch.Tensor([]).to(self.device)

        iter_start = datetime.now()

        input_shape = list(i.shape)
        input_shape[-1] = 1
        i = torch.cat(
            [i, supports.repeat(input_shape)],
            dim=-1)

        o = model(i, hidden0=hidden)
        if type(o) == tuple:
            hidden = o[1]
            o = o[0]
        outputs = torch.cat([outputs, o], dim=1)
        self.logger.log(
            5, "%s -%3d- data shape=(i: %s, o: %s), inputs: %s, outputs: %s, targets: %s" %
            (tag, self.input_sequence-1,
                list(i.shape), list(o.shape),
                list(inputs.shape), list(outputs.shape), list(targets.shape)))

        # 次の入力データ
        i = self.__next_batch_iter_inputs__(i[:, -1:], o[:, -1:])

        for batch_iter in range(self.input_sequence, inputs.shape[1]):
            iter_start = datetime.now()

            # t = targets[:, batch_iter:batch_iter+1]
            input_shape = list(i.shape)
            input_shape[-1] = 1
            i = torch.cat(
                [i, supports.repeat(input_shape)],
                dim=-1)

            o = model(i, hidden0=hidden)
            if type(o) == tuple:
                hidden = o[1]
                o = o[0]
            outputs = torch.cat([outputs, o], dim=1)
            self.logger.log(
                5, "%s -%3d- data shape=(i: %s, o: %s), inputs: %s, outputs: %s, targets: %s" %
                (tag, batch_iter,
                 list(i.shape), list(o.shape),
                 list(inputs.shape), list(outputs.shape), list(targets.shape)))

            # 次の入力データ
            i = self.__next_batch_iter_inputs__(i[:, -1:], o[:, -1:])

        outputs = outputs[:, -targets.shape[1]:]
        loss = criterion(targets, outputs)
        count = np.prod(targets.shape)
        if train:
            loss.backward(retain_graph=True)

        return (loss.cpu().detach().numpy(),
                count,
                outputs.cpu().detach().numpy())

    def train_batch(self,
                    model,
                    criterion,
                    loader,
                    **kwargs):
        total_loss, total_count = 0.0, 0
        if loader.dataset.train:
            loader.dataset.shuffle()
        domain = loader.dataset.domain
        category = loader.dataset.category
        tag = loader.dataset.tag
        loader.epoch += 1

        supports = loader.dataset.support().to(self.device)
        support_features = model.domain_forward(supports)
        if type(support_features) == tuple:
            support_features = support_features[0]

        for batch_no, (inputs, targets) in enumerate(loader):
            start = datetime.now()

            inputs = inputs.to(self.device)
            targets = targets.to(self.device)

            loss, count, _ = self.__inner_batch_iter__(
                model, criterion, inputs, targets, support_features,
                tag=tag,
                epoch=loader.epoch,
                batch_no=batch_no,
                domain=domain,
                category=category,
                train=loader.dataset.train)
            total_loss += loss
            total_count += count

            proc_time = (datetime.now() - start).total_seconds()
            self.logger.debug(
                "%s %04d.%03d %s.%s loss: %f count: %d proc time: %f" %
                (tag, loader.epoch, batch_no, domain, category,
                 total_loss/total_count, total_count, proc_time))
        return total_loss, total_count

    def predict_batch(self,
                      outdir,
                      model,
                      criterion,
                      loader,
                      **kwargs):
        total_loss, total_count = 0.0, 0
        domain = loader.dataset.domain
        category = loader.dataset.category
        tag = loader.dataset.tag
        loader.epoch += 1
        if category is None or len(category) == 0:
            outfile = path.join(outdir, "%s.predict.npy" % (domain))
        else:
            outfile = path.join(outdir,
                                "%s.%s.predict.npy" % (domain, category))

        supports = loader.dataset.support().to(self.device)
        support_features = model.domain_forward(supports)
        if type(support_features) == tuple:
            support_features = support_features[0]

        predicts = []
        for batch_no, (inputs, targets) in enumerate(loader):
            start = datetime.now()

            inputs = inputs.to(self.device)
            targets = targets.to(self.device)

            loss, count, outputs = self.__inner_batch_iter__(
                model, criterion, inputs, targets, support_features,
                tag=tag,
                epoch=loader.epoch,
                batch_no=batch_no,
                domain=domain,
                category=category,
                train=False)
            total_loss += loss
            total_count += count

            self.logger.log(
                5, "%s data shape=(inputs: %s, outputs: %s, targets: %s)" %
                (tag, inputs.shape, outputs.shape, targets.shape))
            # predictデータ保存用にリストに追加
            predicts.append(outputs)

            proc_time = (datetime.now() - start).total_seconds()
            self.logger.debug(
                "%s %04d.%03d %s.%s loss: %f count: %d proc time: %f" %
                (tag, loader.epoch, batch_no, domain, category,
                 total_loss/total_count, total_count, proc_time))

        predicts = np.concatenate(predicts, axis=0)  # batch次元で連結
        np.save(outfile, predicts)  # ファイルに保存

        return total_loss, total_count


class LongTermBidirectionalTrainer(LongTermMetaLearnTrainer):
    def __init__(self,
                 predict_freq: int = 1,
                 num_query: int = 50,
                 num_train_support: int = 0,
                 num_predict_support: int = 0,
                 support_seq_len: int = 5,
                 query_seq_len: int = 5,
                 num_ar_feature: int = 3,
                 use_cuda: bool = False,
                 batch_size: int = 50,
                 workers: int = 1,
                 input_seq: int = 10,
                 delimiter: str = "\t",
                 **kwargs):
        super(LongTermBidirectionalTrainer, self).__init__(
            predict_freq=predict_freq,
            num_query=num_query,
            num_train_support=num_train_support,
            num_predict_support=num_predict_support,
            support_seq_len=support_seq_len,
            query_seq_len=query_seq_len,
            num_ar_feature=num_ar_feature,
            use_cuda=use_cuda,
            batch_size=batch_size,
            workers=workers,
            input_seq=input_seq,
            delimiter=delimiter,
            **kwargs)

    def Model(self, **kwargs):
        """
        input_size: int
        hidden_size: int
        num_layers: int
        domain_input_size: int
        domain_hidden_size: int
        domain_num_layers: int
        """
        # assert "input_size" in kwargs and "hidden_size" in kwargs and "num_layers" in kwargs and "domain_output_size" in kwargs
        # # 元の時系列データの特徴量数に、ドメインの特徴量抽出用モデルの出力次元数を加えたものが入力データ
        assert "input_size" in kwargs and "hidden_size" in kwargs and "num_layers" in kwargs and "domain_output_size" in kwargs
        assert "domain_hidden_size" in kwargs and "domain_num_layers" in kwargs and "domain_output_size" in kwargs
        assert "domain_forward_unit00" in kwargs and "domain_forward_unit01" in kwargs and "domain_forward_unit02" in kwargs
        # 元の時系列データの特徴量数に、ドメインの特徴量抽出用モデルの出力次元数を加えたものが入力データ
        from models import BidrectionalModel as DomainModel
        input_size = int(kwargs["input_size"])
        hidden_size = int(kwargs["hidden_size"])
        num_layers = int(kwargs["num_layers"])
        domain_hidden_size = int(kwargs["domain_hidden_size"])
        domain_num_layers = int(kwargs["domain_num_layers"])
        domain_output_size = int(kwargs["domain_output_size"])
        domain_forward_unit00 = int(kwargs["domain_forward_unit00"])
        domain_forward_unit01 = int(kwargs["domain_forward_unit01"])
        domain_forward_unit02 = int(kwargs["domain_forward_unit02"])
        return DomainModel(input_size,
                           hidden_size,
                           num_layers,
                           output_dim=1,
                           domain_hidden_dim=domain_hidden_size,
                           domain_num_layers=domain_num_layers,
                           domain_output_size=domain_output_size,
                           domain_forward_unit00=domain_forward_unit00,
                           domain_forward_unit01=domain_forward_unit01,
                           domain_forward_unit02=domain_forward_unit02)


class LongTermAttentionTrainer(LongTermMetaLearnTrainer):
    def __init__(self,
                 predict_freq: int = 1,
                 num_query: int = 50,
                 num_train_support: int = 0,
                 num_predict_support: int = 0,
                 support_seq_len: int = 5,
                 query_seq_len: int = 5,
                 num_ar_feature: int = 3,
                 use_cuda: bool = False,
                 batch_size: int = 50,
                 workers: int = 1,
                 input_seq: int = 10,
                 delimiter: str = "\t",
                 **kwargs):
        super(LongTermAttentionTrainer, self).__init__(
            predict_freq=predict_freq,
            num_query=num_query,
            num_train_support=num_train_support,
            num_predict_support=num_predict_support,
            support_seq_len=support_seq_len,
            query_seq_len=query_seq_len,
            num_ar_feature=num_ar_feature,
            use_cuda=use_cuda,
            batch_size=batch_size,
            workers=workers,
            input_seq=input_seq,
            delimiter=delimiter,
            **kwargs)

    def Model(self, **kwargs):
        """
        input_size: int
        hidden_size: int
        num_layers: int
        domain_input_size: int
        domain_hidden_size: int
        domain_num_layers: int
        """
        # assert "input_size" in kwargs and "hidden_size" in kwargs and "num_layers" in kwargs and "domain_output_size" in kwargs
        # # 元の時系列データの特徴量数に、ドメインの特徴量抽出用モデルの出力次元数を加えたものが入力データ
        assert "input_size" in kwargs
        assert "hidden_size" in kwargs
        assert "num_layers" in kwargs
        assert "domain_output_size" in kwargs
        assert "domain_hidden_size" in kwargs
        assert "domain_num_layers" in kwargs
        assert "domain_output_size" in kwargs
        from models import AttentionModel as Model
        # 元の時系列データの特徴量数に、ドメインの特徴量抽出用モデルの出力次元数を加えたものが入力データ
        input_size = int(kwargs["input_size"])
        hidden_size = int(kwargs["hidden_size"])
        num_layers = int(kwargs["num_layers"])
        domain_hidden_size = int(kwargs["domain_hidden_size"])
        domain_num_layers = int(kwargs["domain_num_layers"])
        domain_output_size = int(kwargs["domain_output_size"])
        return Model(input_size,
                     hidden_size,
                     num_layers,
                     output_dim=1,
                     domain_hidden_dim=domain_hidden_size,
                     domain_num_layers=domain_num_layers,
                     domain_output_size=domain_output_size)

    def __inner_batch_iter__(self,
                             model, criterion, inputs, targets, supports,
                             tag: str = "",
                             epoch: int = 0,
                             batch_no: int = 0,
                             domain: str = "",
                             category: str = "",
                             train: bool = False):
        hidden = None
        i = inputs[:, 0:self.input_sequence]
        outputs = torch.Tensor([]).to(self.device)

        iter_start = datetime.now()

        self.logger.log(
            5, "%s -%3d- data shape=(i: %s, o: %s, s: %s), inputs: %s, outputs: %s, targets: %s" %
            (tag, self.input_sequence-1,
                list(i.shape), None, list(supports.shape),
                list(inputs.shape), list(outputs.shape), list(targets.shape)))

        o = model(i, supports, hidden0=hidden)
        if type(o) == tuple:
            hidden = o[1]
            o = o[0]
        outputs = torch.cat([outputs, o], dim=1)
        self.logger.log(
            5, "%s -%3d- data shape=(i: %s, o: %s), inputs: %s, outputs: %s, targets: %s" %
            (tag, self.input_sequence-1,
                list(i.shape), list(o.shape),
                list(inputs.shape), list(outputs.shape), list(targets.shape)))

        # 次の入力データ
        # start_idx = self.input_sequence + batch_iter
        i = self.__next_batch_iter_inputs__(i[:, -1:], o[:, -1:])

        for batch_iter in range(self.input_sequence, inputs.shape[1]):
            iter_start = datetime.now()

            self.logger.log(
                5, "%s -%3d- data shape=(i: %s, o: %s, s: %s), inputs: %s, outputs: %s, targets: %s" %
                (tag, batch_iter,
                 list(i.shape), None, list(supports.shape),
                 list(inputs.shape), list(outputs.shape), list(targets.shape)))

            o = model(i, supports, hidden0=hidden)
            if type(o) == tuple:
                hidden = o[1]
                o = o[0]
            outputs = torch.cat([outputs, o], dim=1)
            self.logger.log(
                5, "%s -%3d- data shape=(i: %s, o: %s), inputs: %s, outputs: %s, targets: %s" %
                (tag, batch_iter,
                 list(i.shape), list(o.shape),
                 list(inputs.shape), list(outputs.shape), list(targets.shape)))

            # 次の入力データ
            # start_idx = self.input_sequence + batch_iter
            i = self.__next_batch_iter_inputs__(i[:, -1:], o[:, -1:])

        outputs = outputs[:, -targets.shape[1]:]
        loss = criterion(targets, outputs)
        count = np.prod(targets.shape)
        if train:
            loss.backward(retain_graph=True)

        return (loss.cpu().detach().numpy(),
                count,
                outputs.cpu().detach().numpy())
