#! -*- coding: utf-8
import logging
import os
import sys
from datetime import datetime
from os import path

import numpy as np
import torch

from datasets import LinearDataset as LinearDataset
from datasets import LstmDataset as Dataset

from .AbstractLongTermTrainer import AbstractLongTermTrainer as AbstractTrainer

__all__ = ["LongTermLstmTrainer",
           "LongTermLinearTrainer",
           "LongTermNonLinearTrainer"]


class LongTermLstmTrainer(AbstractTrainer):
    def __init__(self,
                 predict_freq: int = 1,
                 num_query: int = 50,
                 num_train_support: int = 0,
                 num_predict_support: int = 0,
                 support_seq_len: int = 5,
                 query_seq_len: int = 5,
                 num_ar_feature: int = 3,
                 use_cuda: bool = False,
                 batch_size: int = 50,
                 workers: int = 1,
                 input_seq: int = 10,
                 delimiter: str = "\t",
                 **kwargs):
        super(LongTermLstmTrainer, self).__init__(
            predict_freq=predict_freq,
            num_query=num_query,
            num_train_support=num_train_support,
            num_predict_support=num_predict_support,
            support_seq_len=support_seq_len,
            query_seq_len=query_seq_len,
            num_ar_feature=num_ar_feature,
            use_cuda=use_cuda,
            batch_size=batch_size,
            workers=workers,
            input_seq=input_seq,
            delimiter=delimiter,
            **kwargs)

    def Model(self, **kwargs):
        """
        input_size: int
        hidden_size: int
        num_layers: int
        """
        assert "input_size" in kwargs and "hidden_size" in kwargs and "num_layers" in kwargs
        input_size = int(kwargs["input_size"])
        hidden_size = int(kwargs["hidden_size"])
        num_layers = int(kwargs["num_layers"])

        from models import LstmModel as Model
        return Model(input_size, hidden_size, num_layers)


class LongTermLinearTrainer(AbstractTrainer):
    def __init__(self,
                 predict_freq: int = 1,
                 num_query: int = 50,
                 num_train_support: int = 0,
                 num_predict_support: int = 0,
                 support_seq_len: int = 5,
                 query_seq_len: int = 5,
                 num_ar_feature: int = 3,
                 use_cuda: bool = False,
                 batch_size: int = 50,
                 workers: int = 1,
                 input_seq: int = 10,
                 delimiter: str = "\t",
                 **kwargs):
        super(LongTermLinearTrainer, self).__init__(
            predict_freq=predict_freq,
            num_query=num_query,
            num_train_support=num_train_support,
            num_predict_support=num_predict_support,
            support_seq_len=support_seq_len,
            query_seq_len=query_seq_len,
            num_ar_feature=num_ar_feature,
            use_cuda=use_cuda,
            batch_size=batch_size,
            workers=workers,
            input_seq=input_seq,
            delimiter=delimiter,
            **kwargs)

    def Model(self, **kwargs):
        """
        input_size: int
        """
        assert "input_size" in kwargs
        from models import LinearModel as Model
        input_size = self.num_ar_feature
        output_size = int(kwargs["input_size"])
        return Model(input_size, output_size)

    def Dataset(self, queryfile, supportfile,
                tag: str = "",
                num_query: int = None,
                num_support: int = 3,
                query_seq_len: int = None,
                train: bool = False,
                **kwargs):
        return LinearDataset(queryfile, supportfile,
                             tag=tag,
                             predict_freq=self.predict_freq,
                             num_query=num_query,
                             num_support=num_support,
                             support_seq_len=self.support_seq_len,
                             query_seq_len=query_seq_len,
                             num_ar_feature=self.num_ar_feature,
                             train=train)

    def __next_batch_iter_inputs__(self, inputs, outputs):
        """ 次回batch iterationで使用する入力データを作成する
        =====
        今回入力したデータの特徴量次元末尾に今回の出力データを追加。
        追加したデータ数分、特徴量次元の先頭データを除外する。

        =====
        inputs: tensor
            今回入力データ
        outputs: tensor
            今回出力データ
        =====
        next_inputs: tensor
            次回入力データ
        """
        i = torch.cat([inputs, outputs], dim=2)
        i = i[:, :, outputs.shape[-1]:]
        return i


class LongTermNonLinearTrainer(LongTermLinearTrainer):
    def __init__(self,
                 predict_freq: int = 1,
                 num_query: int = 50,
                 num_train_support: int = 0,
                 num_predict_support: int = 0,
                 support_seq_len: int = 5,
                 query_seq_len: int = 5,
                 num_ar_feature: int = 3,
                 use_cuda: bool = False,
                 batch_size: int = 50,
                 workers: int = 1,
                 input_seq: int = 10,
                 delimiter: str = "\t",
                 **kwargs):
        super(LongTermNonLinearTrainer, self).__init__(
            predict_freq=predict_freq,
            num_query=num_query,
            num_train_support=num_train_support,
            num_predict_support=num_predict_support,
            support_seq_len=support_seq_len,
            query_seq_len=query_seq_len,
            num_ar_feature=num_ar_feature,
            use_cuda=use_cuda,
            batch_size=batch_size,
            workers=workers,
            input_seq=input_seq,
            delimiter=delimiter,
            **kwargs)

    def Model(self, **kwargs):
        """
        input_size: int
        hidden_size: int
        """
        assert "input_size" in kwargs and "hidden_size" in kwargs
        from models import NonLinearModel as Model
        input_size = self.num_ar_feature
        hidden_size = int(kwargs["hidden_size"])
        output_size = int(kwargs["input_size"])
        return Model(input_size, output_size, hidden_size)
